# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import azure.cosmos.documents as documents
from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
from azure.cosmos.partition_key import PartitionKey
import urllib3

import asyncio
import config

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = "index-samples"
PARTITION_KEY = PartitionKey(path='/id', kind='Hash')

# A typical container has the following properties within it's indexingPolicy property
#   indexingMode
#   automatic
#   includedPaths
#   excludedPaths
#
# We can toggle 'automatic' to either be True or False depending upon whether we want to have indexing over all columns by default or not.
#
# We can provide options while creating documents. indexingDirective is one such,
# by which we can tell whether it should be included or excluded in the index of the parent container.
# indexingDirective can be either 'Include', 'Exclude' or 'Default'


# To run this Demo, please provide your own CA certs file or download one from
#     http://curl.haxx.se/docs/caextract.html
# Setup the certificate file in .pem format.
# If you still get an SSLError, try disabling certificate verification and suppress warnings

find_entity_by_id_query = {
        "query": "SELECT * FROM r WHERE r.id=@id",
        "parameters": [
            { "name":"@id", "value": id }
        ]
    }

def obtain_client():
    # Try to setup the cacert.pem
    # connection_policy.SSLConfiguration.SSLCaCerts = CaCertPath
    # Else, disable verification
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    return CosmosClient(HOST, MASTER_KEY)


# Query for Entity / Entities
async def query_entities(parent, entity_type, id = None):
    find_entity_by_id_query = {
            "query": "SELECT * FROM r WHERE r.id=@id",
            "parameters": [
                { "name":"@id", "value": id }
            ]
        }
    entities = None
    try:
        if entity_type == 'database':
            if id == None:
                entities = [entity async for entity in parent.list_databases()]
            else:
                entities = [entity async for entity in parent.query_databases(find_entity_by_id_query)]

        elif entity_type == 'container':
            if id == None:
                entities = [entity async for entity in parent.list_containers()]
            else:
                entities = [entity async for entity in parent.query_containers(find_entity_by_id_query)]

        elif entity_type == 'document':
            if id == None:
                entities = [entity async for entity in parent.read_all_items()]
            else:
                entities = [entity async for entity in parent.query_items(find_entity_by_id_query)]
    except exceptions.AzureError as e:
        print("The following error occurred while querying for the entity / entities ", entity_type, id if id != None else "")
        print(e)
        raise
    if id == None:
        return entities
    if entities and len(entities) == 1:
        return entities[0]
    return None


async def delete_container_if_exists(db, container_id):
    try:
        await db.delete_container(container_id)
        print('Container with id \'{0}\' was deleted'.format(container_id))
    except exceptions.CosmosResourceNotFoundError:
        pass
    except exceptions.CosmosHttpResponseError as e:
        if e.status_code == 400:
            print("Bad request for container link", container_id)
        raise


def print_dictionary_items(dict):
    for k, v in dict.items():
        print("{:<15}".format(k), v)
    print()


async def fetch_all_databases(client):
    databases = await query_entities(client, 'database')
    print("-" * 41)
    print("-" * 41)
    for db in databases:
        print_dictionary_items(db)
        print("-" * 41)


async def query_documents_with_custom_query(container, query_with_optional_parameters, message = "Document(s) found by query: "):
    try:
        results = container.query_items(query_with_optional_parameters)
        print(message)
        async for doc in results:
            print(doc)
        return results
    except exceptions.CosmosResourceNotFoundError:
        print("Document doesn't exist")
    except exceptions.CosmosHttpResponseError as e:
        if e.status_code == 400:
            # Can occur when we are trying to query on excluded paths
            print("Bad Request exception occurred: ", e)
            pass
        else:
            raise
    finally:
        print()


async def explicitly_exclude_from_index(db):
    """ The default index policy on a DocumentContainer will AUTOMATICALLY index ALL documents added.
        There may be scenarios where you want to exclude a specific doc from the index even though all other
        documents are being indexed automatically.
        This method demonstrates how to use an index directive to control this

    """
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Create a container with default index policy (i.e. automatic = true)
        created_Container = await db.create_container(id=CONTAINER_ID, partition_key=PARTITION_KEY)
        print(created_Container)

        print("\n" + "-" * 25 + "\n1. Container created with index policy")
        properties = await created_Container.read()
        print_dictionary_items(properties["indexingPolicy"])

        # Create a document and query on it immediately.
        # Will work as automatic indexing is still True
        doc = await created_Container.create_item(body={ "id" : "doc1", "orderId" : "order1" })
        print("\n" + "-" * 25 + "Document doc1 created with order1" +  "-" * 25)
        print(doc)

        query = {
                "query": "SELECT * FROM r WHERE r.orderId=@orderNo",
                "parameters": [ { "name":"@orderNo", "value": "order1" } ]
            }
        await query_documents_with_custom_query(created_Container, query)

        # Now, create a document but this time explicitly exclude it from the container using IndexingDirective
        # Then query for that document
        # Should NOT find it, because we excluded it from the index
        # BUT, the document is there and doing a ReadDocument by Id will prove it
        doc2 = await created_Container.create_item(
            body={ "id" : "doc2", "orderId" : "order2" },
            indexing_directive=documents.IndexingDirective.Exclude
        )
        print("\n" + "-" * 25 + "Document doc2 created with order2" +  "-" * 25)
        print(doc2)

        query = {
                "query": "SELECT * FROM r WHERE r.orderId=@orderNo",
                "parameters": [ { "name":"@orderNo", "value": "order2" } ]
                }
        await query_documents_with_custom_query(created_Container, query)

        docRead = await created_Container.read_item(item="doc2", partition_key="doc2")
        print("Document read by ID: \n", docRead["id"])

        # Cleanup
        await db.delete_container(created_Container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def use_manual_indexing(db):
    """The default index policy on a DocumentContainer will AUTOMATICALLY index ALL documents added.
       There may be cases where you can want to turn-off automatic indexing and only selectively add only specific documents to the index.
       This method demonstrates how to control this by setting the value of automatic within indexingPolicy to False

    """
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Create a container with manual (instead of automatic) indexing
        created_Container = await db.create_container(
            id=CONTAINER_ID,
            indexing_policy={"automatic" : False},
            partition_key=PARTITION_KEY
        )
        properties = await created_Container.read()
        print(created_Container)

        print("\n" + "-" * 25 + "\n2. Container created with index policy")
        print_dictionary_items(properties["indexingPolicy"])

        # Create a document
        # Then query for that document
        # We should find nothing, because automatic indexing on the container level is False
        # BUT, the document is there and doing a ReadDocument by Id will prove it
        doc = await created_Container.create_item(body={ "id" : "doc1", "orderId" : "order1" })
        print("\n" + "-" * 25 + "Document doc1 created with order1" +  "-" * 25)
        print(doc)

        query = {
                "query": "SELECT * FROM r WHERE r.orderId=@orderNo",
                "parameters": [ { "name":"@orderNo", "value": "order1" } ]
            }
        await query_documents_with_custom_query(created_Container, query)

        docRead = await created_Container.read_item(item="doc1", partition_key="doc1")
        print("Document read by ID: \n", docRead["id"])

        # Now create a document, passing in an IndexingDirective saying we want to specifically index this document
        # Query for the document again and this time we should find it because we manually included the document in the index
        doc2 = await created_Container.create_item(
            body={ "id" : "doc2", "orderId" : "order2" },
            indexing_directive=documents.IndexingDirective.Include
        )
        print("\n" + "-" * 25 + "Document doc2 created with order2" +  "-" * 25)
        print(doc2)

        query = {
                "query": "SELECT * FROM r WHERE r.orderId=@orderNo",
                "parameters": [ { "name":"@orderNo", "value": "order2" } ]
            }
        await query_documents_with_custom_query(created_Container, query)

        # Cleanup
        await db.delete_container(created_Container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def exclude_paths_from_index(db):
    """The default behavior is for Cosmos to index every attribute in every document automatically.
       There are times when a document contains large amounts of information, in deeply nested structures
       that you know you will never search on. In extreme cases like this, you can exclude paths from the
       index to save on storage cost, improve write performance and also improve read performance because the index is smaller

       This method demonstrates how to set excludedPaths within indexingPolicy
    """
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        doc_with_nested_structures = {
            "id" : "doc1",
            "foo" : "bar",
            "metaData" : "meta",
            "subDoc" : { "searchable" : "searchable", "nonSearchable" : "value" },
            "excludedNode" : { "subExcluded" : "something",  "subExcludedNode" : { "someProperty" : "value" } }
            }
        container_to_create = { "id" : CONTAINER_ID ,
                                "indexingPolicy" :
                                {
                                    "includedPaths" : [ {'path' : "/*"} ], # Special mandatory path of "/*" required to denote include entire tree
                                    "excludedPaths" : [ {'path' : "/metaData/*"}, # exclude metaData node, and anything under it
                                                        {'path' : "/subDoc/nonSearchable/*"}, # exclude ONLY a part of subDoc
                                                        {'path' : "/\"excludedNode\"/*"} # exclude excludedNode node, and anything under it
                                                      ]
                                    }
                                }
        print(container_to_create)
        print(doc_with_nested_structures)
        # Create a container with the defined properties
        # The effect of the above IndexingPolicy is that only id, foo, and the subDoc/searchable are indexed
        created_Container = await db.create_container(
            id=container_to_create['id'],
            indexing_policy=container_to_create['indexingPolicy'],
            partition_key=PARTITION_KEY
        )
        properties = await created_Container.read()
        print(created_Container)
        print("\n" + "-" * 25 + "\n4. Container created with index policy")
        print_dictionary_items(properties["indexingPolicy"])

        # The effect of the above IndexingPolicy is that only id, foo, and the subDoc/searchable are indexed
        doc = await created_Container.create_item(body=doc_with_nested_structures)
        print("\n" + "-" * 25 + "Document doc1 created with nested structures" +  "-" * 25)
        print(doc)

        # Querying for a document on either metaData or /subDoc/subSubDoc/someProperty > fail because these paths were excluded and they raise a BadRequest(400) Exception
        query = {"query": "SELECT * FROM r WHERE r.metaData=@desiredValue", "parameters" : [{ "name":"@desiredValue", "value": "meta" }]}
        await query_documents_with_custom_query(created_Container, query)

        query = {"query": "SELECT * FROM r WHERE r.subDoc.nonSearchable=@desiredValue", "parameters" : [{ "name":"@desiredValue", "value": "value" }]}
        await query_documents_with_custom_query(created_Container, query)

        query = {"query": "SELECT * FROM r WHERE r.excludedNode.subExcludedNode.someProperty=@desiredValue", "parameters" : [{ "name":"@desiredValue", "value": "value" }]}
        await query_documents_with_custom_query(created_Container, query)

        # Querying for a document using foo, or even subDoc/searchable > succeed because they were not excluded
        query = {"query": "SELECT * FROM r WHERE r.foo=@desiredValue", "parameters" : [{ "name":"@desiredValue", "value": "bar" }]}
        await query_documents_with_custom_query(created_Container, query)

        query = {"query": "SELECT * FROM r WHERE r.subDoc.searchable=@desiredValue", "parameters" : [{ "name":"@desiredValue", "value": "searchable" }]}
        await query_documents_with_custom_query(created_Container, query)

        # Cleanup
        await db.delete_container(created_Container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def range_scan_on_hash_index(db):
    """When a range index is not available (i.e. Only hash or no index found on the path), comparisons queries can still
       be performed as scans using Allow scan request headers passed through options

       This method demonstrates how to force a scan when only hash indexes exist on the path

       ===== Warning=====
       This was made an opt-in model by design.
       Scanning is an expensive operation and doing this will have a large impact
       on RequestUnits charged for an operation and will likely result in queries being throttled sooner.
    """
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Force a range scan operation on a hash indexed path
        container_to_create = { "id" : CONTAINER_ID ,
                                "indexingPolicy" :
                                {
                                    "includedPaths" : [ {'path' : "/"} ],
                                    "excludedPaths" : [ {'path' : "/length/*"} ] # exclude length
                                    }
                                }
        created_Container = await db.create_container(
            id=container_to_create['id'],
            indexing_policy=container_to_create['indexingPolicy'],
            partition_key=PARTITION_KEY
        )
        properties = await created_Container.read()
        print(created_Container)
        print("\n" + "-" * 25 + "\n5. Container created with index policy")
        print_dictionary_items(properties["indexingPolicy"])

        doc1 = await created_Container.create_item(body={ "id" : "dyn1", "length" : 10, "width" : 5, "height" : 15 })
        doc2 = await created_Container.create_item(body={ "id" : "dyn2", "length" : 7, "width" : 15 })
        doc3 = await created_Container.create_item(body={ "id" : "dyn3", "length" : 2 })
        print("Three docs created with ids : ", doc1["id"], doc2["id"], doc3["id"])

        # Query for length > 5 - fail, this is a range based query on a Hash index only document
        query = { "query": "SELECT * FROM r WHERE r.length > 5" }
        await query_documents_with_custom_query(created_Container, query)

        # Now add IndexingDirective and repeat query
        # expect 200 OK because now we are explicitly allowing scans in a query
        # using the enableScanInQuery directive
        results = created_Container.query_items(
            query,
            enable_scan_in_query=True
        )
        print("Printing documents queried by range by providing enableScanInQuery = True")
        async for doc in results: print(doc["id"])

        # Cleanup
        await db.delete_container(created_Container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def use_range_indexes_on_strings(db):
    """Showing how range queries can be performed even on strings.

    """
    try:
        await delete_container_if_exists(db, CONTAINER_ID)
        # containers = query_entities(client, 'container', parent_link = database_link)
        # print(containers)

        # Use range indexes on strings

        # This is how you can specify a range index on strings (and numbers) for all properties.
        # This is the recommended indexing policy for containers. i.e. precision -1
        #indexingPolicy = {
        #    'indexingPolicy': {
        #        'includedPaths': [
        #            {
        #                'indexes': [
        #                    {
        #                        'kind': documents.IndexKind.Range,
        #                        'dataType': documents.DataType.String,
        #                        'precision': -1
        #                    }
        #                ]
        #            }
        #        ]
        #    }
        #}

        # For demo purposes, we are going to use the default (range on numbers, hash on strings) for the whole document (/* )
        # and just include a range index on strings for the "region".
        container_definition = {
            'id': CONTAINER_ID,
            'indexingPolicy': {
                'includedPaths': [
                    {
                        'path': '/region/?',
                        'indexes': [
                            {
                                'kind': documents.IndexKind.Range,
                                'dataType': documents.DataType.String,
                                'precision': -1
                            }
                        ]
                    },
                    {
                        'path': '/*'
                    }
                ]
            }
        }

        created_Container = await db.create_container(
            id=container_definition['id'],
            indexing_policy=container_definition['indexingPolicy'],
            partition_key=PARTITION_KEY
        )
        properties = await created_Container.read()
        print(created_Container)
        print("\n" + "-" * 25 + "\n6. Container created with index policy")
        print_dictionary_items(properties["indexingPolicy"])

        await created_Container.create_item(body={ "id" : "doc1", "region" : "USA" })
        await created_Container.create_item(body={ "id" : "doc2", "region" : "UK" })
        await created_Container.create_item(body={ "id" : "doc3", "region" : "Armenia" })
        await created_Container.create_item(body={ "id" : "doc4", "region" : "Egypt" })

        # Now ordering against region is allowed. You can run the following query
        query = { "query" : "SELECT * FROM r ORDER BY r.region" }
        message = "Documents ordered by region"
        await query_documents_with_custom_query(created_Container, query, message)

        # You can also perform filters against string comparison like >= 'UK'. Note that you can perform a prefix query,
        # the equivalent of LIKE 'U%' (is >= 'U' AND < 'U')
        query = { "query" : "SELECT * FROM r WHERE r.region >= 'U'" }
        message = "Documents with region begining with U"
        await query_documents_with_custom_query(created_Container, query, message)

        # Cleanup
        await db.delete_container(created_Container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def perform_index_transformations(db):
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Create a container with default indexing policy
        created_Container = await db.create_container(id=CONTAINER_ID, partition_key=PARTITION_KEY)
        properties = await created_Container.read()
        print(created_Container)

        print("\n" + "-" * 25 + "\n7. Container created with index policy")
        print_dictionary_items(properties["indexingPolicy"])

        # Insert some documents
        doc1 = await created_Container.create_item(body={ "id" : "dyn1", "length" : 10, "width" : 5, "height" : 15 })
        doc2 = await created_Container.create_item(body={ "id" : "dyn2", "length" : 7, "width" : 15 })
        doc3 = await created_Container.create_item(body={ "id" : "dyn3", "length" : 2 })
        print("Three docs created with ids : ", doc1["id"], doc2["id"], doc3["id"], " with indexing mode", properties['indexingPolicy']['indexingMode'])

        # Switch to use string & number range indexing with maximum precision.
        print("Changing to string & number range indexing with maximum precision (needed for Order By).")

        properties['indexingPolicy']['includedPaths'][0]['indexes'] = [{
            'kind': documents.IndexKind.Range,
            'dataType': documents.DataType.String,
            'precision': -1
        }]

        created_Container = await db.replace_container(
            container=created_Container.id,
            partition_key=PARTITION_KEY,
            indexing_policy=properties['indexingPolicy']
        )
        properties = await created_Container.read()

        # Check progress and wait for completion - should be instantaneous since we have only a few documents, but larger
        # containers will take time.
        print_dictionary_items(properties["indexingPolicy"])

        # Now exclude a path from indexing to save on storage space.
        print("Now excluding the path /length/ to save on storage space")
        properties['indexingPolicy']['excludedPaths'] = [{"path" : "/length/*"}]

        created_Container = await db.replace_container(
            container=created_Container.id,
            partition_key=PARTITION_KEY,
            indexing_policy=properties['indexingPolicy']
        )
        properties = await created_Container.read()
        print_dictionary_items(properties["indexingPolicy"])

        # Cleanup
        await db.delete_container(created_Container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def perform_multi_orderby_query(db):
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Create a container with composite indexes
        indexing_policy = {
            "compositeIndexes": [
                [
                    {
                        "path": "/numberField",
                        "order": "ascending"
                    },
                    {
                        "path": "/stringField",
                        "order": "descending"
                    }
                ],
                [
                    {
                        "path": "/numberField",
                        "order": "descending"
                    },
                    {
                        "path": "/stringField",
                        "order": "ascending"
                    },
                    {
                        "path": "/numberField2",
                        "order": "descending"
                    },
                    {
                        "path": "/stringField2",
                        "order": "ascending"
                    }
                ]
            ]
        }

        created_container = await db.create_container(
            id=CONTAINER_ID,
            indexing_policy=indexing_policy,
            partition_key=PARTITION_KEY
        )
        print(created_container)
        properties = await created_container.read()

        print("\n" + "-" * 25 + "\n8. Container created with index policy")
        print_dictionary_items(properties["indexingPolicy"])

        # Insert some documents
        await created_container.create_item(body={"id": "doc1", "numberField": 1, "stringField": "1", "numberField2": 1, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc2", "numberField": 1, "stringField": "1", "numberField2": 1, "stringField2": "2"})
        await created_container.create_item(body={"id": "doc3", "numberField": 1, "stringField": "1", "numberField2": 2, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc4", "numberField": 1, "stringField": "1", "numberField2": 2, "stringField2": "2"})
        await created_container.create_item(body={"id": "doc5", "numberField": 1, "stringField": "2", "numberField2": 1, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc6", "numberField": 1, "stringField": "2", "numberField2": 1, "stringField2": "2"})
        await created_container.create_item(body={"id": "doc7", "numberField": 1, "stringField": "2", "numberField2": 2, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc8", "numberField": 1, "stringField": "2", "numberField2": 2, "stringField2": "2"})
        await created_container.create_item(body={"id": "doc9", "numberField": 2, "stringField": "1", "numberField2": 1, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc10", "numberField": 2, "stringField": "1", "numberField2": 1, "stringField2": "2"})
        await created_container.create_item(body={"id": "doc11", "numberField": 2, "stringField": "1", "numberField2": 2, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc12", "numberField": 2, "stringField": "1", "numberField2": 2, "stringField2": "2"})
        await created_container.create_item(body={"id": "doc13", "numberField": 2, "stringField": "2", "numberField2": 1, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc14", "numberField": 2, "stringField": "2", "numberField2": 1, "stringField2": "2"})
        await created_container.create_item(body={"id": "doc15", "numberField": 2, "stringField": "2", "numberField2": 2, "stringField2": "1"})
        await created_container.create_item(body={"id": "doc16", "numberField": 2, "stringField": "2", "numberField2": 2, "stringField2": "2"})

        print("Query documents and Order by 1st composite index: Ascending numberField and Descending stringField:")

        query = {
                "query": "SELECT * FROM r ORDER BY r.numberField ASC, r.stringField DESC",
                }
        await query_documents_with_custom_query(created_container, query)

        print("Query documents and Order by inverted 2nd composite index -")
        print("Ascending numberField, Descending stringField, Ascending numberField2, Descending stringField2")

        query = {
                "query": "SELECT * FROM r ORDER BY r.numberField ASC, r.stringField DESC, r.numberField2 ASC, r.stringField2 DESC",
                }
        await query_documents_with_custom_query(created_container, query)

        # Cleanup
        await db.delete_container(created_container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def use_geospatial_indexing_policy(db):
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Create a container with geospatial indexes
        indexing_policy = {
            'includedPaths': [
                {'path': '/"Location"/?',
                    'indexes': [
                        {
                            'kind': 'Spatial',
                            'dataType': 'Point'
                        }]
                 },
                {
                    'path': '/'
                }
            ]
        }

        created_container = await db.create_container(
            id=CONTAINER_ID,
            partition_key=PARTITION_KEY,
            indexing_policy=indexing_policy
        )
        properties = await created_container.read()
        print(created_container)

        print("\n" + "-" * 25 + "\n9. Container created with geospatial indexes")
        print_dictionary_items(properties["indexingPolicy"])

        # Create some items
        doc9 = await created_container.create_item(body={"id": "loc1", 'Location': {'type': 'Point', 'coordinates': [20.0, 20.0]}})
        doc9 = await created_container.create_item(body={"id": "loc2", 'Location': {'type': 'Point', 'coordinates': [100.0, 100.0]}})

        # Run ST_DISTANCE queries using the geospatial index
        query = "SELECT * FROM root WHERE (ST_DISTANCE(root.Location, {type: 'Point', coordinates: [20.1, 20]}) < 20000)"
        await query_documents_with_custom_query(created_container, query)

        # Cleanup
        await db.delete_container(created_container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def use_vector_embedding_policy(db):
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Create a container with vector embedding policy and vector indexes
        indexing_policy = {
            "vectorIndexes": [
                {"path": "/vector", "type": "quantizedFlat", "quantizationByteSize": 8},
                {"path": "/vector2", "type": "diskANN", "vectorIndexShardKey": ["/city"], "indexingSearchListSize": 50}
            ]
        }
        vector_embedding_policy = {
            "vectorEmbeddings": [
                {
                    "path": "/vector",
                    "dataType": "float32",
                    "dimensions": 256,
                    "distanceFunction": "euclidean"
                },
                {
                    "path": "/vector2",
                    "dataType": "int8",
                    "dimensions": 200,
                    "distanceFunction": "dotproduct"
                }
            ]
        }

        created_container = await db.create_container(
            id=CONTAINER_ID,
            partition_key=PARTITION_KEY,
            indexing_policy=indexing_policy,
            vector_embedding_policy=vector_embedding_policy
        )
        properties = await created_container.read()
        print(created_container)

        print("\n" + "-" * 25 + "\n10. Container created with vector embedding policy and vector indexes")
        print_dictionary_items(properties["indexingPolicy"])
        print_dictionary_items(properties["vectorEmbeddingPolicy"])

        # We define our own get_embeddings() function for the sake of the sample, but you should be using a third
        # party service to generate these properly.
        def get_embeddings(num):
            return [num, num, num]

        # Create some items with vector embeddings
        for i in range(10):
            created_container.create_item({"id": "vector_item" + str(i), "embeddings": get_embeddings(i)})

        # Run vector similarity search queries using VectorDistance
        query = "SELECT TOP 5 c.id,VectorDistance(c.embeddings, [{}]) AS " \
                "SimilarityScore FROM c ORDER BY VectorDistance(c.embeddings, [{}])".format(get_embeddings(1),
                                                                                            get_embeddings(1))
        await query_documents_with_custom_query(created_container, query)

        # Run vector similarity search queries using VectorDistance with specifications
        query = "SELECT TOP 5 c.id,VectorDistance(c.embeddings, [{}], true, {{'dataType': 'float32' ," \
                " 'distanceFunction': 'cosine'}}) AS SimilarityScore FROM c ORDER BY VectorDistance(c.embeddings," \
                " [{}], true, {{'dataType': 'float32', 'distanceFunction': 'cosine'}})".format(get_embeddings(1),
                                                                                               get_embeddings(1))
        await query_documents_with_custom_query(created_container, query)

        # Cleanup
        await db.delete_container(created_container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def use_full_text_policy(db):
    try:
        await delete_container_if_exists(db, CONTAINER_ID)

        # Create a container with full text policy and full text indexes
        indexing_policy = {
            "automatic": True,
            "fullTextIndexes": [
                {"path": "/text1"}
            ]
        }
        full_text_policy = {
            "defaultLanguage": "en-US",
            "fullTextPaths": [
                {
                    "path": "/text1",
                    "language": "en-US"
                },
                {
                    "path": "/text2",
                    "language": "en-US"
                }
            ]
        }

        created_container = await db.create_container(
            id=CONTAINER_ID,
            partition_key=PARTITION_KEY,
            indexing_policy=indexing_policy,
            full_text_policy=full_text_policy
        )
        properties = await created_container.read()
        print(created_container)

        print("\n" + "-" * 25 + "\n11. Container created with full text policy and full text indexes")
        print_dictionary_items(properties["indexingPolicy"])
        print_dictionary_items(properties["fullTextPolicy"])

        # Create some items to use with full text search
        sample_texts = ["Common popular pop music artists include Taylor Swift and The Weekend.",
                        "The weekend is coming up soon, do you have any plans?",
                        "Depending on the artist, their music can be very different.",
                        "Mozart and Beethoven are some of the most recognizable names in classical music.",
                        "Taylor acts in many movies, and is considered a great artist."]

        # Create some items to use with full text search
        for i in range(5):
            created_container.create_item({"id": "full_text_item" + str(i), "text1": sample_texts[i],
                                           "vector": [1, 2, 3]})

        # Run full text search queries using full text score ranking
        query = "SELECT TOP 3 c.text1 FROM c ORDER BY RANK FullTextScore(c.text1, ['artist']))"
        await query_documents_with_custom_query(created_container, query)

        # Run full text search queries using RRF ranking
        query = "SELECT TOP 3 c.text1 FROM c ORDER BY RANK RRF(FullTextScore(c.text1, ['music'])))"
        await query_documents_with_custom_query(created_container, query)

        # Run hybrid search queries using RRF ranking wth vector distances
        query = "SELECT TOP 3 c.text1 FROM c ORDER BY RANK RRF(FullTextScore(c.text1, ['music'])," \
                " VectorDistance(c.vector, [1, 2, 3]))"
        await query_documents_with_custom_query(created_container, query)

        # Cleanup
        db.delete_container(created_container)
        print("\n")
    except exceptions.CosmosResourceExistsError:
        print("Entity already exists")
    except exceptions.CosmosResourceNotFoundError:
        print("Entity doesn't exist")


async def run_sample():
    try:
        async with obtain_client() as client:
            await fetch_all_databases(client)

            # Create database if it doesn't exist already.
            created_db = await client.create_database_if_not_exists(DATABASE_ID)
            print(created_db)

            # 1. Exclude a document from the index
            await explicitly_exclude_from_index(created_db)

            # 2. Use manual (instead of automatic) indexing
            await use_manual_indexing(created_db)

            # 4. Exclude specified document paths from the index
            await exclude_paths_from_index(created_db)

            # 5. Force a range scan operation on a hash indexed path
            await range_scan_on_hash_index(created_db)

            # 6. Use range indexes on strings
            await use_range_indexes_on_strings(created_db)

            # 7. Perform an index transform
            await perform_index_transformations(created_db)

            # 8. Perform Multi Orderby queries using composite indexes
            await perform_multi_orderby_query(created_db)

            # 9. Create and use a geospatial indexing policy
            await use_geospatial_indexing_policy(created_db)

            # 10. Create and use a vector embedding policy
            await use_vector_embedding_policy(created_db)

            # 11. Create and use a full text policy
            await use_full_text_policy(created_db)

    except exceptions.AzureError as e:
        raise e

if __name__ == '__main__':
    asyncio.run(run_sample())



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest

import azure.cosmos._base as base
import azure.cosmos.cosmos_client as cosmos_client
import test_config
from azure.cosmos._execution_context import base_execution_context as base_execution_context
from azure.cosmos.exceptions import CosmosHttpResponseError
from azure.cosmos.partition_key import PartitionKey


def get_database_link(database):
    return 'dbs/' + database.id


def get_document_collection_link(database, document_collection):
    return get_database_link(database) + '/colls/' + document_collection.id


@pytest.mark.cosmosEmulator
class TestQueryExecutionContextEndToEnd(unittest.TestCase):
    """Routing Map Functionalities end-to-end Tests.
    """

    created_collection = None
    document_definitions = None
    created_db = None
    client: cosmos_client.CosmosClient = None
    config = test_config.TestConfig
    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    connectionPolicy = test_config.TestConfig.connectionPolicy
    TEST_DATABASE_ID = config.TEST_DATABASE_ID

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

        cls.client = cosmos_client.CosmosClient(cls.host, cls.masterKey)
        cls.created_db = cls.client.get_database_client(cls.TEST_DATABASE_ID)
        cls.created_collection = cls.created_db.create_container(
            id='query_execution_context_tests_' + str(uuid.uuid4()),
            partition_key=PartitionKey(path='/id', kind='Hash')
        )
        cls.document_definitions = []

        # create a document using the document definition
        for i in range(20):
            d = {'id': str(i),
                 'name': 'sample document',
                 'spam': 'eggs' + str(i),
                 'key': 'value'}
            cls.document_definitions.append(d)
        cls.insert_doc(cls.document_definitions)

    @classmethod
    def tearDownClass(cls):
        try:
            cls.created_db.delete_container(cls.created_collection.id)
        except CosmosHttpResponseError:
            pass

    def setUp(self):
        # sanity check:
        partition_key_ranges = list(self.client.client_connection._ReadPartitionKeyRanges(
            get_document_collection_link(self.created_db, self.created_collection)))
        self.assertGreaterEqual(len(partition_key_ranges), 1)

        # sanity check: read documents after creation
        queried_docs = list(self.created_collection.read_all_items())
        self.assertEqual(
            len(queried_docs),
            len(self.document_definitions),
            'create should increase the number of documents')

    def test_no_query_default_execution_context(self):

        options = {'maxItemCount': 2}

        self._test_default_execution_context(options, None, 20)

    def test_no_query_default_execution_context_with_small_last_page(self):

        options = {'maxItemCount': 3}

        self._test_default_execution_context(options, None, 20)

    def test_simple_query_default_execution_context(self):

        query = {
            'query': 'SELECT * FROM root r WHERE r.id != @id',
            'parameters': [
                {'name': '@id', 'value': '5'}
            ]
        }

        options = {'enableCrossPartitionQuery': True, 'maxItemCount': 2}

        res = self.created_collection.query_items(
            query=query,
            enable_cross_partition_query=True,
            max_item_count=2
        )
        self.assertEqual(len(list(res)), 19)

        self._test_default_execution_context(options, query, 19)

    def test_simple_query_default_execution_context_with_small_last_page(self):

        query = {
            'query': 'SELECT * FROM root r WHERE r.id != @id',
            'parameters': [
                {'name': '@id', 'value': '5'}
            ]
        }

        options = {}
        options['enableCrossPartitionQuery'] = True
        options['maxItemCount'] = 3

        self._test_default_execution_context(options, query, 19)

    def _test_default_execution_context(self, options, query, expected_number_of_results):

        page_size = options['maxItemCount']
        collection_link = get_document_collection_link(self.created_db, self.created_collection)
        path = base.GetPathFromLink(collection_link, 'docs')
        collection_id = base.GetResourceIdOrFullNameFromLink(collection_link)

        def fetch_fn(options):
            return self.client.client_connection.QueryFeed(path,
                                                           collection_id,
                                                           query,
                                                           options)

        ######################################
        # test next() behavior
        ######################################
        ex = base_execution_context._DefaultQueryExecutionContext(self.client.client_connection, options, fetch_fn)

        it = ex.__iter__()

        def invokeNext():
            return next(it)

        results = {}
        # validate that invocations of next() produces the same results as expected
        for _ in range(expected_number_of_results):
            item = invokeNext()
            results[item['id']] = item

        self.assertEqual(len(results), expected_number_of_results)

        # after the result set is exhausted, invoking next must raise a StopIteration exception
        self.assertRaises(StopIteration, invokeNext)

        ######################################
        # test fetch_next_block() behavior
        ######################################
        ex = base_execution_context._DefaultQueryExecutionContext(self.client.client_connection, options, fetch_fn)

        results = {}
        cnt = 0
        fetched_res = ex.fetch_next_block()
        fetched_size = 0
        while fetched_res is not None and len(fetched_res) > 0:
            fetched_size = len(fetched_res)

            for item in fetched_res:
                results[item['id']] = item
            cnt += fetched_size
            fetched_res = ex.fetch_next_block()

        # validate the number of collected results
        self.assertEqual(len(results), expected_number_of_results)

        self.assertTrue(fetched_size > 0, "fetched size is 0")
        self.assertTrue(fetched_size <= page_size, "fetched page size greater than page size")

        # no more results will be returned
        self.assertEqual(ex.fetch_next_block(), [])

    @classmethod
    def insert_doc(cls, document_definitions):
        # create a document using the document definition
        created_docs = []
        for d in document_definitions:
            created_doc = cls.created_collection.create_item(body=d)
            created_docs.append(created_doc)

        return created_docs


if __name__ == "__main__":
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import os
import unittest
import uuid
from time import sleep

import pytest
from azure.core.exceptions import ServiceResponseError, ServiceRequestError

import test_config
from azure.cosmos import _partition_health_tracker, _location_cache
from azure.cosmos import CosmosClient
from azure.cosmos._partition_health_tracker import HEALTH_STATUS, UNHEALTHY_TENTATIVE, UNHEALTHY
from azure.cosmos.exceptions import CosmosHttpResponseError
from _fault_injection_transport import FaultInjectionTransport
from test_per_partition_circuit_breaker_mm import create_doc, write_operations_and_errors, operations, REGION_1, \
    REGION_2, PK_VALUE, perform_write_operation, perform_read_operation, CREATE, READ, validate_stats

COLLECTION = "created_collection"

def validate_unhealthy_partitions(global_endpoint_manager,
                                  expected_unhealthy_partitions):
    health_info_map = global_endpoint_manager.global_partition_endpoint_manager_core.partition_health_tracker.pk_range_wrapper_to_health_info
    unhealthy_partitions = 0
    for pk_range_wrapper, location_to_health_info in health_info_map.items():
        for location, health_info in location_to_health_info.items():
            health_status = health_info.unavailability_info.get(HEALTH_STATUS)
            if health_status == UNHEALTHY_TENTATIVE or health_status == UNHEALTHY:
                unhealthy_partitions += 1

            else:
                assert health_info.read_consecutive_failure_count < 10
            # single region write account should never track write failures
            assert health_info.write_failure_count == 0
            assert health_info.write_consecutive_failure_count == 0

    assert unhealthy_partitions == expected_unhealthy_partitions

@pytest.mark.cosmosCircuitBreakerMultiRegion
class TestPerPartitionCircuitBreakerSmMrr:
    host = test_config.TestConfig.host
    master_key = test_config.TestConfig.masterKey
    connectionPolicy = test_config.TestConfig.connectionPolicy
    TEST_DATABASE_ID = test_config.TestConfig.TEST_DATABASE_ID
    TEST_CONTAINER_MULTI_PARTITION_ID = test_config.TestConfig.TEST_MULTI_PARTITION_CONTAINER_ID

    def setup_method_with_custom_transport(self, custom_transport, default_endpoint=host, **kwargs):
        container_id = kwargs.pop("container_id", None)
        if not container_id:
            container_id = self.TEST_CONTAINER_MULTI_PARTITION_ID
        client = CosmosClient(default_endpoint, self.master_key, consistency_level="Session",
                              preferred_locations=[REGION_1, REGION_2],
                              transport=custom_transport, **kwargs)
        db = client.get_database_client(self.TEST_DATABASE_ID)
        container = db.get_container_client(container_id)
        return {"client": client, "db": db, "col": container}

    def setup_info(self, error, **kwargs):
        expected_uri = _location_cache.LocationCache.GetLocationalEndpoint(self.host, REGION_2)
        uri_down = _location_cache.LocationCache.GetLocationalEndpoint(self.host, REGION_1)
        custom_transport = FaultInjectionTransport()
        # two documents targeted to same partition, one will always fail and the other will succeed
        doc = create_doc()
        predicate = lambda r: (FaultInjectionTransport.predicate_is_document_operation(r) and
                               FaultInjectionTransport.predicate_targets_region(r, uri_down))
        custom_transport.add_fault(predicate,
                                   error)
        custom_setup = self.setup_method_with_custom_transport(custom_transport, default_endpoint=self.host, **kwargs)
        setup = self.setup_method_with_custom_transport(None, default_endpoint=self.host, **kwargs)
        return setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate

    def test_stat_reset(self):
        error_lambda = lambda r: FaultInjectionTransport.error_after_delay(
            0,
            CosmosHttpResponseError(
                status_code=503,
                message="Some injected error.")
        )
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = \
            self.setup_info(error_lambda, container_id=test_config.TestConfig.TEST_SINGLE_PARTITION_CONTAINER_ID)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        container.upsert_item(body=doc)
        sleep(1)
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager
        # lower refresh interval for testing
        _partition_health_tracker.REFRESH_INTERVAL_MS = 10 * 1000
        try:
            for i in range(2):
                validate_unhealthy_partitions(global_endpoint_manager, 0)
                # read will fail and retry in other region
                perform_read_operation(READ,
                                       fault_injection_container,
                                       doc['id'],
                                       PK_VALUE,
                                       expected_uri)
                try:
                    perform_write_operation(CREATE,
                                            container,
                                            fault_injection_container,
                                            str(uuid.uuid4()),
                                            PK_VALUE,
                                            expected_uri)
                except CosmosHttpResponseError as e:
                    assert e.status_code == 503
            validate_unhealthy_partitions(global_endpoint_manager, 0)
            validate_stats(global_endpoint_manager, 0,  2, 2, 0, 0, 0)
            sleep(25)
            perform_read_operation(READ,
                                   fault_injection_container,
                                   doc['id'],
                                   PK_VALUE,
                                   expected_uri)

            validate_stats(global_endpoint_manager, 0, 3, 1, 0, 0, 0)
        finally:
            _partition_health_tracker.REFRESH_INTERVAL_MS = 60 * 1000

    @pytest.mark.parametrize("write_operation, error", write_operations_and_errors())
    def test_write_consecutive_failure_threshold(self, write_operation, error):
        error_lambda = lambda r: FaultInjectionTransport.error_after_delay(0, error)
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        global_endpoint_manager = container.client_connection._global_endpoint_manager

        # writes should fail in sm mrr with circuit breaker and should not mark unavailable a partition
        for i in range(6):
            validate_unhealthy_partitions(global_endpoint_manager, 0)
            with pytest.raises((CosmosHttpResponseError, ServiceResponseError)):
                perform_write_operation(
                    write_operation,
                    container,
                    fault_injection_container,
                    str(uuid.uuid4()),
                    PK_VALUE,
                    expected_uri,
                )

        validate_unhealthy_partitions(global_endpoint_manager, 0)

    @pytest.mark.parametrize("write_operation, error", write_operations_and_errors())
    def test_write_failure_rate_threshold(self, write_operation, error):
        error_lambda = lambda r: FaultInjectionTransport.error_after_delay(0, error)
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager
        # lower minimum requests for testing
        _partition_health_tracker.MINIMUM_REQUESTS_FOR_FAILURE_RATE = 10
        os.environ["AZURE_COSMOS_FAILURE_PERCENTAGE_TOLERATED"] = "80"
        try:
            # writes should fail but still be tracked and mark unavailable a partition after crossing threshold
            for i in range(10):
                validate_unhealthy_partitions(global_endpoint_manager, 0)
                if i == 4 or i == 8:
                    # perform some successful creates to reset consecutive counter
                    # remove faults and perform a write
                    custom_transport.faults = []
                    fault_injection_container.upsert_item(body=doc)
                    custom_transport.add_fault(predicate,
                                               lambda r: FaultInjectionTransport.error_after_delay(
                                                   0,
                                                   error
                                               ))
                else:
                    with pytest.raises((CosmosHttpResponseError, ServiceResponseError)) as exc_info:
                        perform_write_operation(write_operation,
                                                      container,
                                                      fault_injection_container,
                                                      str(uuid.uuid4()),
                                                      PK_VALUE,
                                                      expected_uri)
                    assert exc_info.value == error

            validate_unhealthy_partitions(global_endpoint_manager, 0)

        finally:
            os.environ["AZURE_COSMOS_FAILURE_PERCENTAGE_TOLERATED"] = "90"
            # restore minimum requests
            _partition_health_tracker.MINIMUM_REQUESTS_FOR_FAILURE_RATE = 100

    @pytest.mark.parametrize("read_operation, write_operation", operations())
    def test_service_request_error(self, read_operation, write_operation):
        # the region should be tried 4 times before failing over and mark the region as unavailable
        # the partition should not be marked as unavailable
        error_lambda = lambda r: FaultInjectionTransport.error_region_down()
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        container.upsert_item(body=doc)
        perform_read_operation(read_operation,
                                     fault_injection_container,
                                     doc['id'],
                                     PK_VALUE,
                                     expected_uri)
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

        # there shouldn't be partition marked as unavailable
        validate_unhealthy_partitions(global_endpoint_manager, 0)
        assert len(global_endpoint_manager.location_cache.location_unavailability_info_by_endpoint) == 1

        # recover partition
        # remove faults and reduce initial recover time and perform a write
        original_unavailable_time = _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS
        _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = 1
        custom_transport.faults = []
        try:
            perform_read_operation(read_operation,
                                         fault_injection_container,
                                         doc['id'],
                                         PK_VALUE,
                                         expected_uri)
        finally:
            _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = original_unavailable_time
        validate_unhealthy_partitions(global_endpoint_manager, 0)

        custom_transport.add_fault(predicate,
                                   lambda r: FaultInjectionTransport.error_region_down())

        # for single write region account if regional endpoint is down there will be write unavailability
        with pytest.raises(ServiceRequestError):
            perform_write_operation(write_operation,
                                          container,
                                          fault_injection_container,
                                          str(uuid.uuid4()),
                                          PK_VALUE,
                                          expected_uri)

        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

        validate_unhealthy_partitions(global_endpoint_manager, 0)
        # there shouldn't be region marked as unavailable
        assert len(global_endpoint_manager.location_cache.location_unavailability_info_by_endpoint) == 1

    # test cosmos client timeout

if __name__ == '__main__':
    unittest.main()




# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Service-specific Exceptions in the Azure Cosmos database service.
"""
from azure.core.exceptions import (
    AzureError,
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError
)
from . import http_constants
from .http_constants import StatusCodes as _StatusCode, SubStatusCodes as _SubStatusCodes

class CosmosHttpResponseError(HttpResponseError):
    """An HTTP request to the Azure Cosmos database service has failed."""

    def __init__(self, status_code=None, message=None, response=None, **kwargs):
        """
        :param int status_code: HTTP response code.
        :param str message: Error message.
        """
        self.headers = response.headers if response else {}
        self.sub_status = kwargs.pop('sub_status', None)
        self.http_error_message = message
        status = status_code or (int(response.status_code) if response else 0)

        if http_constants.HttpHeaders.SubStatus in self.headers:
            self.sub_status = int(self.headers[http_constants.HttpHeaders.SubStatus])
            formatted_message = "Status code: %d Sub-status: %d\n%s" % (status, self.sub_status, str(message))
        else:
            formatted_message = "Status code: %d\n%s" % (status, str(message))

        super(CosmosHttpResponseError, self).__init__(message=formatted_message, response=response, **kwargs)
        self.status_code = status


class CosmosResourceNotFoundError(ResourceNotFoundError, CosmosHttpResponseError):
    """An HTTP error response with status code 404."""


class CosmosResourceExistsError(ResourceExistsError, CosmosHttpResponseError):
    """An HTTP error response with status code 409."""


class CosmosAccessConditionFailedError(CosmosHttpResponseError):
    """An HTTP error response with status code 412."""


class CosmosBatchOperationError(HttpResponseError):
    """A transactional batch request to the Azure Cosmos database service has failed.

    :ivar int error_index: Index of operation within the batch that caused the error.
    :ivar headers: Error headers.
    :vartype headers: dict[str, Any]
    :ivar status_code: HTTP response code.
    :vartype status_code: int
    :ivar message: Error message.
    :vartype message: str
    :ivar operation_responses: List of failed operations' responses.
    :vartype operation_responses: List[dict[str, Any]]

    .. admonition:: Example:

        .. literalinclude:: ../samples/document_management.py
            :start-after: [START handle_batch_error]
            :end-before: [END handle_batch_error]
            :language: python
            :dedent: 0
            :caption: Handle a CosmosBatchOperationError:
            :name: handle_batch_error
    """

    def __init__(
            self,
            error_index=None,
            headers=None,
            status_code=None,
            message=None,
            operation_responses=None,
            **kwargs):
        self.error_index = error_index
        self.headers = headers
        self.sub_status = None
        self.http_error_message = message
        self.operation_responses = operation_responses
        status = status_code

        if http_constants.HttpHeaders.SubStatus in self.headers:
            self.sub_status = int(self.headers[http_constants.HttpHeaders.SubStatus])
            formatted_message = "Status code: %d Sub-status: %d\n%s" % (status, self.sub_status, str(message))
        else:
            formatted_message = "Status code: %d\n%s" % (status, str(message))

        super(CosmosBatchOperationError, self).__init__(message=formatted_message, response=None, **kwargs)
        self.status_code = status


class CosmosClientTimeoutError(AzureError):
    """An operation failed to complete within the specified timeout."""

    def __init__(self, **kwargs):
        message = kwargs.pop('message', None)
        if message is None:
            message = "The request failed to complete within the given timeout."
        self.response = None
        self.history = None
        super(CosmosClientTimeoutError, self).__init__(message, **kwargs)

class _InternalCosmosException:
    def __init__(self, status_code, headers, reason=None):
        self.status_code = status_code
        self.headers = headers
        self.reason = reason

def _partition_range_is_gone(e):
    if (e.status_code == _StatusCode.GONE
            and e.sub_status == _SubStatusCodes.PARTITION_KEY_RANGE_GONE):
        return True
    return False

def _container_recreate_exception(e) -> bool:
    is_bad_request = e.status_code == _StatusCode.BAD_REQUEST
    is_collection_rid_mismatch = e.sub_status == _SubStatusCodes.COLLECTION_RID_MISMATCH

    is_not_found = e.status_code == _StatusCode.NOT_FOUND
    is_throughput_not_found = e.sub_status == _SubStatusCodes.THROUGHPUT_OFFER_NOT_FOUND

    return (is_bad_request and is_collection_rid_mismatch) or (is_not_found and is_throughput_not_found)

def _is_partition_split_or_merge(e):
    return e.status_code == _StatusCode.GONE and e.sub_status == _SubStatusCodes.COMPLETING_SPLIT



# The MIT License (MIT)
# Copyright (c) 2021 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for partition key range cache implementation in the Azure
Cosmos database service.
"""
import logging
from typing import Any, Optional

from ... import _base
from ..collection_routing_map import CollectionRoutingMap
from .. import routing_range

_LOGGER = logging.getLogger(__name__)

# pylint: disable=protected-access


class PartitionKeyRangeCache(object):
    """
    PartitionKeyRangeCache provides list of effective partition key ranges for a
    collection.

    This implementation loads and caches the collection routing map per
    collection on demand.
    """

    def __init__(self, client):
        """
        Constructor
        """

        self._documentClient = client

        # keeps the cached collection routing map by collection id
        self._collection_routing_map_by_item = {}

    async def get_overlapping_ranges(self, collection_link, partition_key_ranges, feed_options, **kwargs):
        """Given a partition key range and a collection, return the list of
        overlapping partition key ranges.

        :param str collection_link: The name of the collection.
        :param list partition_key_ranges: List of partition key range.
        :param dict feed_options: The request options.
        :return: List of overlapping partition key ranges.
        :rtype: list
        """
        collection_id = _base.GetResourceIdOrFullNameFromLink(collection_link)
        pk_range_options = _base.format_pk_range_options(feed_options)
        await self.init_collection_routing_map_if_needed(collection_link, collection_id, pk_range_options, **kwargs)

        return self._collection_routing_map_by_item[collection_id].get_overlapping_ranges(partition_key_ranges)

    async def init_collection_routing_map_if_needed(
            self,
            collection_link: str,
            collection_id: str,
            feed_options: dict[str, Any],
            **kwargs: dict[str, Any]
    ):
        collection_routing_map = self._collection_routing_map_by_item.get(collection_id)
        if collection_routing_map is None:
            # Pass _internal_pk_range_fetch flag to prevent recursive 410 retry logic
            # When a 410 partition split error occurs, the SDK calls refresh_routing_map_provider()
            # which clears the cache and retries. The retry needs partition key ranges, which calls
            # this method, which triggers _ReadPartitionKeyRanges. If that query also goes through
            # the 410 retry logic and calls refresh again, we get infinite recursion.
            _LOGGER.debug(
                "PK range cache (async): Initializing routing map for collection_id=%s with "
                "_internal_pk_range_fetch=True to prevent recursive 410 retry.",
                collection_id
            )
            pk_range_kwargs = {**kwargs, "_internal_pk_range_fetch": True}
            collection_pk_ranges = [pk async for pk in
                                    self._documentClient._ReadPartitionKeyRanges(collection_link,
                                                                                 feed_options,
                                                                                 **pk_range_kwargs)]
            # for large collections, a split may complete between the read partition key ranges query page responses,
            # causing the partitionKeyRanges to have both the children ranges and their parents. Therefore, we need
            # to discard the parent ranges to have a valid routing map.
            collection_pk_ranges = list(PartitionKeyRangeCache._discard_parent_ranges(collection_pk_ranges))
            collection_routing_map = CollectionRoutingMap.CompleteRoutingMap(
                [(r, True) for r in collection_pk_ranges], collection_id
            )
            self._collection_routing_map_by_item[collection_id] = collection_routing_map
            _LOGGER.debug(
                "PK range cache (async): Cached routing map for collection_id=%s with %d ranges",
                collection_id, len(collection_pk_ranges)
            )

    async def get_range_by_partition_key_range_id(
            self,
            collection_link: str,
            partition_key_range_id: int,
            feed_options: dict[str, Any],
            **kwargs: dict[str, Any]
    ) -> Optional[dict[str, Any]]:
        collection_id = _base.GetResourceIdOrFullNameFromLink(collection_link)
        pk_range_options = _base.format_pk_range_options(feed_options)
        await self.init_collection_routing_map_if_needed(collection_link, collection_id, pk_range_options, **kwargs)

        return self._collection_routing_map_by_item[collection_id].get_range_by_partition_key_range_id(
            partition_key_range_id)

    @staticmethod
    def _discard_parent_ranges(partitionKeyRanges):
        parentIds = set()
        for r in partitionKeyRanges:
            if isinstance(r, dict) and routing_range.PartitionKeyRange.Parents in r:
                for parentId in r[routing_range.PartitionKeyRange.Parents]:
                    parentIds.add(parentId)
        return (r for r in partitionKeyRanges if r[routing_range.PartitionKeyRange.Id] not in parentIds)


def _second_range_is_after_first_range(range1, range2):
    if range1.max > range2.min:
        ##r.min < #previous_r.max
        return False

    if range2.min == range1.max and range1.isMaxInclusive and range2.isMinInclusive:
        # the inclusive ending endpoint of previous_r is the same as the inclusive beginning endpoint of r
        return False

    return True


def _is_sorted_and_non_overlapping(ranges):
    for idx, r in list(enumerate(ranges))[1:]:
        previous_r = ranges[idx - 1]
        if not _second_range_is_after_first_range(previous_r, r):
            return False
    return True


def _subtract_range(r, partition_key_range):
    """Evaluates and returns r - partition_key_range

    :param dict partition_key_range: Partition key range.
    :param routing_range.Range r: query range.
    :return: The subtract r - partition_key_range.
    :rtype: routing_range.Range
    """

    left = max(partition_key_range[routing_range.PartitionKeyRange.MaxExclusive], r.min)

    if left == r.min:
        leftInclusive = r.isMinInclusive
    else:
        leftInclusive = False

    queryRange = routing_range.Range(left, r.max, leftInclusive, r.isMaxInclusive)
    return queryRange


class SmartRoutingMapProvider(PartitionKeyRangeCache):
    """
    Efficiently uses PartitionKeyRangeCache and minimizes the unnecessary
    invocation of CollectionRoutingMap.get_overlapping_ranges()
    """

    async def get_overlapping_ranges(self, collection_link, partition_key_ranges, feed_options = None, **kwargs):
        """
        Given the sorted ranges and a collection,
        Returns the list of overlapping partition key ranges

        :param str collection_link: The collection link.
        :param (list of routing_range.Range) partition_key_ranges:
            The sorted list of non-overlapping ranges.
        :param dict feed_options: The request options.
        :return: List of partition key ranges.
        :rtype: list of dict
        :raises ValueError:
            If two ranges in partition_key_ranges overlap or if the list is not sorted
        """

        # validate if the list is non-overlapping and sorted
        if not _is_sorted_and_non_overlapping(partition_key_ranges):
            raise ValueError("the list of ranges is not a non-overlapping sorted ranges")

        target_partition_key_ranges = []

        it = iter(partition_key_ranges)
        try:
            currentProvidedRange = next(it)
            while True:
                if currentProvidedRange.isEmpty():
                    # skip and go to the next item\
                    currentProvidedRange = next(it)
                    continue

                if target_partition_key_ranges:
                    queryRange = _subtract_range(currentProvidedRange, target_partition_key_ranges[-1])
                else:
                    queryRange = currentProvidedRange

                overlappingRanges =\
                    await PartitionKeyRangeCache.get_overlapping_ranges(
                        self,
                        collection_link,
                        [queryRange],
                        feed_options,
                        **kwargs)
                assert overlappingRanges, "code bug: returned overlapping ranges for queryRange {} is empty".format(
                    queryRange
                )
                target_partition_key_ranges.extend(overlappingRanges)

                lastKnownTargetRange = routing_range.Range.PartitionKeyRangeToRange(target_partition_key_ranges[-1])

                # the overlapping ranges must contain the requested range
                assert (
                    currentProvidedRange.max <= lastKnownTargetRange.max
                ), "code bug: returned overlapping ranges {} does not contain the requested range {}".format(
                    overlappingRanges, queryRange
                )

                # the current range is contained in target_partition_key_ranges just move forward
                currentProvidedRange = next(it)

                while currentProvidedRange.max <= lastKnownTargetRange.max:
                    # the current range is covered too. just move forward
                    currentProvidedRange = next(it)
        except StopIteration:
            # when the iteration is exhausted we get here. There is nothing else to be done
            pass

        return target_partition_key_ranges



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest
from azure.core.exceptions import AzureError

from azure.cosmos.exceptions import CosmosResourceNotFoundError, CosmosHttpResponseError

import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.partition_key as partition_key
import test_config
from azure.cosmos import DatabaseProxy, PartitionKey, ContainerProxy
from typing import Any, Dict, Union

ItemDict = Dict[str, Any]
PkField = Union[str, None, type[partition_key.NonePartitionKeyValue], type[partition_key.NullPartitionKeyValue]]
COLLECTION = "created_collection"
DATABASE = "created_db"
VERSIONS = [1, 2]


def _new_null_pk_doc(pk_field: PkField) -> ItemDict:
    if pk_field is None:
        return {'id': str(uuid.uuid4())}
    return {'id': str(uuid.uuid4()), pk_field: None}


def _assert_pk(item: ItemDict, pk_field: PkField = 'pk') -> None:
    if pk_field:
        assert item[pk_field] is None
    else:
        assert pk_field not in item


def _read_and_assert(container: ContainerProxy, doc_id: str, pk_field: PkField = 'pk', pk_value: Any = None) -> None:
    item = container.read_item(item=doc_id, partition_key=pk_value)
    _assert_pk(item, pk_field)


def _query_and_assert_single(container: ContainerProxy, pk_field: PkField = 'pk', pk_value: Any = None) -> None:
    query_iterable = container.query_items(query='SELECT * from c', partition_key=pk_value)
    results = list(query_iterable)
    assert len(results) > 0
    for result in results:
        _assert_pk(result, pk_field)


def _change_feed_and_assert(container: ContainerProxy, pk_field: PkField = 'pk', pk_value: Any = None) -> None:
    changes_iterable = container.query_items_change_feed(is_start_from_beginning=True, partition_key=pk_value)
    changes = list(changes_iterable)
    assert len(changes) > 0
    for change in changes:
        _assert_pk(change, pk_field)


def _replace_update_cycle(container: ContainerProxy, doc: ItemDict, pk_field: PkField = 'pk') -> None:
    doc['newField'] = 'newValue'
    item = container.replace_item(item=doc['id'], body=doc)
    assert item['newField'] == 'newValue'
    _assert_pk(item, pk_field)
    item = container.upsert_item(body=doc)
    _assert_pk(item, pk_field)


def _patch_and_assert(container: ContainerProxy, doc_id: str, pk_field: PkField = 'pk', pk_value: Any = None) -> None:
    operations = [{"op": "add", "path": "/favorite_color", "value": "red"}]
    item = container.patch_item(item=doc_id, partition_key=pk_value, patch_operations=operations)
    assert item['favorite_color'] == 'red'
    _assert_pk(item, pk_field)


def _batch_and_assert(container: ContainerProxy, pk_field: PkField = 'pk', pk_value: Any = None) -> None:
    if pk_value == partition_key.NullPartitionKeyValue:
        batch_operations = [
            ("upsert", (({"id": str(uuid.uuid4()), pk_field: None}),)),
        ]
    elif pk_field:
        batch_operations = [
            ("upsert", (({"id": str(uuid.uuid4()), pk_field: pk_value}),)),
        ]
    else:
        batch_operations = [
            ("upsert", (({"id": str(uuid.uuid4())}),)),
        ]
    items = container.execute_item_batch(batch_operations=batch_operations, partition_key=pk_value)
    assert len(items) == 1
    _assert_pk(items[0]['resourceBody'], pk_field)

@pytest.fixture(scope="class")
def setup():
    if (TestPartitionKey.masterKey == '[YOUR_KEY_HERE]' or
            TestPartitionKey.host == '[YOUR_ENDPOINT_HERE]'):
        raise Exception(
            "You must specify your Azure Cosmos account values for "
            "'masterKey' and 'host' at the top of this class to run the "
            "tests.")
    test_client = cosmos_client.CosmosClient(TestPartitionKey.host, TestPartitionKey.masterKey),
    created_db = test_client[0].get_database_client(TestPartitionKey.TEST_DATABASE_ID)
    return {
        DATABASE: created_db,
        COLLECTION: created_db.get_container_client(TestPartitionKey.TEST_CONTAINER_ID)
    }


def _assert_no_conflicts(container: ContainerProxy, pk_value: PkField) -> None:
    conflict_definition: ItemDict = {'id': 'new conflict', 'resourceId': 'doc1', 'operationType': 'create', 'resourceType': 'document'}
    if pk_value:
        assert len(list(container.query_conflicts(query='SELECT * FROM root r WHERE r.resourceType=\'' + conflict_definition.get('resourceType') + '\'', partition_key=pk_value))) == 0
    else:
        try:
            list(container.query_conflicts(query='SELECT * FROM root r WHERE r.resourceType=\'' + conflict_definition.get('resourceType') + '\'', partition_key=pk_value))
            pytest.fail("Should have thrown since cross partition query not allowed")
        except CosmosHttpResponseError as ex:
            assert ex.status_code, 400
    for api_call in (container.get_conflict, container.delete_conflict):
        try:
            api_call(conflict_definition['id'], partition_key=pk_value)
            pytest.fail("Should have thrown since there are no conflicts")
        except AzureError as ex:
            assert isinstance(ex, CosmosResourceNotFoundError)


def _perform_operations_on_pk(created_container, pk_field, pk_value):
    # Create initial null PK doc
    document_definition = _new_null_pk_doc(pk_field)
    item = created_container.create_item(body=document_definition)
    _assert_pk(item, pk_field)
    # Read & point reads
    _read_and_assert(created_container, document_definition['id'], pk_field, pk_value)
    items_iter = created_container.read_items(items=[(document_definition['id'], pk_value)])
    items_list = list(items_iter)
    assert len(items_list) == 1
    _assert_pk(items_list[0], pk_field)
    # Change feed
    # If partition key is missing, change feed does not work
    if pk_field:
        _change_feed_and_assert(created_container, pk_field, pk_value)
    # Query
    if pk_value:
        _query_and_assert_single(created_container, pk_field, pk_value)
    else:
        try:
            # passing None directly should throw as requires cross partition param
            _query_and_assert_single(created_container, pk_field, pk_value)
            pytest.fail("Should have thrown")
        except CosmosHttpResponseError as ex:
            assert ex.status_code, 400
    # Replace / Upsert cycle
    _replace_update_cycle(created_container, document_definition, pk_field)
    # Patch
    _patch_and_assert(created_container, document_definition['id'], pk_field, pk_value)
    # Batch
    _batch_and_assert(created_container, pk_field, pk_value)
    # Conflicts
    _assert_no_conflicts(created_container, pk_value)
    # Get Feed Range
    feed_range = created_container.feed_range_from_partition_key(partition_key=pk_value)
    assert feed_range is not None
    items_iter = created_container.query_items(query='SELECT * FROM c',
                                               feed_range=feed_range,
                                               enable_cross_partition_query=True)
    items_list = list(items_iter)
    assert len(items_list) > 0
    for item in items_list:
        _assert_pk(item, pk_field)
    # Delete
    created_container.delete_item(item=document_definition['id'], partition_key=pk_value)
    # recreate the item to test delete item api
    created_container.create_item(body=document_definition)
    created_container.delete_all_items_by_partition_key(partition_key=pk_value)


@pytest.mark.cosmosEmulator
@pytest.mark.unittest
@pytest.mark.usefixtures("setup")
class TestPartitionKey:
    """Tests to verify if non-partitioned collections are properly accessed on migration with version 2018-12-31.
    """

    client: cosmos_client.CosmosClient = None
    created_db: DatabaseProxy = None
    created_collection: ContainerProxy = None
    host: str = test_config.TestConfig.host
    masterKey: str = test_config.TestConfig.masterKey
    connectionPolicy = test_config.TestConfig.connectionPolicy  # policy object (SDK internal type)
    TEST_DATABASE_ID: str = test_config.TestConfig.TEST_DATABASE_ID
    TEST_CONTAINER_ID: str = test_config.TestConfig.TEST_MULTI_PARTITION_CONTAINER_ID

    @pytest.mark.parametrize("version", VERSIONS)
    def test_multi_partition_collection_read_document_with_no_pk(self, setup, version) -> None:
        pk_val: PkField = partition_key.NonePartitionKeyValue  # type: ignore[assignment]
        created_container = setup[DATABASE].create_container_if_not_exists(
            id="container_with_no_pk" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk", kind='Hash', version=version)
        )
        try:
           _perform_operations_on_pk(created_container, None, pk_val)
        finally:
            setup[DATABASE].delete_container(created_container)

    @pytest.mark.parametrize("version", VERSIONS)
    def test_with_null_pk(self, setup, version) -> None:
        pk_field = 'pk'
        pk_vals = [None, partition_key.NullPartitionKeyValue]
        created_container = setup[DATABASE].create_container_if_not_exists(
            id="container_with_nul_pk" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk", kind='Hash', version=version)
        )
        try:
            for pk_value in pk_vals:
                _perform_operations_on_pk(created_container, pk_field, pk_value)
        finally:
            setup[DATABASE].delete_container(created_container)

    def test_hash_v2_partition_key_definition(self, setup) -> None:
        created_container_properties = setup[COLLECTION].read()
        assert created_container_properties['partitionKey']['version'] == 2

    def test_hash_v1_partition_key_definition(self, setup) -> None:
        created_container = setup[DATABASE].create_container(
            id='container_with_pkd_v2' + str(uuid.uuid4()),
            partition_key=partition_key.PartitionKey(path="/id", kind="Hash", version=1)
        )
        created_container_properties = created_container.read()
        assert created_container_properties['partitionKey']['version'] == 1
        setup[DATABASE].delete_container(created_container)


if __name__ == '__main__':
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.exceptions as exceptions
from azure.cosmos.partition_key import PartitionKey
import azure.cosmos.documents as documents


import config
import json
from typing import Dict, Any

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://learn.microsoft.com/azure/cosmos-db/create-sql-api-python#create-a-database-account
#
# 2. Microsoft Azure Cosmos
#    pip install azure-cosmos>=4.0.0
# ----------------------------------------------------------------------------------------------------------
# Sample - how to get and use resource token that allows restricted access to data
# ----------------------------------------------------------------------------------------------------------
# Note:
#
# This sample creates a Container to your database account.
# Each time a Container is created the account will be billed for 1 hour of usage based on
# the provisioned throughput (RU/s) of that account.
# ----------------------------------------------------------------------------------------------------------
# Adding region name to use the code sample in docs 
#<configureConnectivity>
HOST = config.settings["host"]
MASTER_KEY = config.settings["master_key"]

DATABASE_ID = config.settings["database_id"]
CONTAINER_ID = config.settings["container_id"]
PARTITION_KEY = PartitionKey(path="/username")


# User that you want to give access to
USERNAME, USERNAME_2 = "user", "user2"

CONTAINER_ALL_PERMISSION = "CONTAINER_ALL_PERMISSION"
PARTITION_READ_PERMISSION = "PARTITION_READ_PERMISSION"
DOCUMENT_ALL_PERMISSION = "DOCUMENT_ALL_PERMISSION"


def create_user_if_not_exists(db, username):
    try:
        user = db.create_user(body={"id": username})
    except exceptions.CosmosResourceExistsError:
        user = db.get_user_client(username)

    return user


def create_permission_if_not_exists(user, permission_definition):
    try:
        permission = user.create_permission(permission_definition)
    except exceptions.CosmosResourceExistsError:
        permission = user.get_permission(permission_definition["id"])

    return permission


def token_client_upsert(container, username, item_id):
    try:
        container.upsert_item(
            {
                "id": item_id,
                "username": username,
                "msg": "This is a message for " + username,
            }
        )
    except exceptions.CosmosHttpResponseError:
        print("Error in upserting item with id '{0}'.".format(item_id))


def token_client_read_all(container):
    try:
        items = list(container.read_all_items())
        for i in items:
            print(i)
    except exceptions.CosmosResourceNotFoundError:
        print("Cannot read items--container '{0}' not found.".format(container.id))
    except exceptions.CosmosHttpResponseError:
        print("Error in reading items in container '{0}'.".format(container.id))


def token_client_read_item(container, username, item_id):
    try:
        item = container.read_item(item=item_id, partition_key=username)
        print(item)
    except exceptions.CosmosResourceNotFoundError:
        print("Cannot read--item with id '{0}' not found.".format(item_id))
    except exceptions.CosmosHttpResponseError:
        print("Error in reading item with id '{0}'.".format(item_id))


def token_client_delete(container, username, item_id):
    try:
        container.delete_item(item=item_id, partition_key=username)
    except exceptions.CosmosResourceNotFoundError:
        print("Cannot delete--item with id '{0}' not found.".format(item_id))
    except exceptions.CosmosHttpResponseError:
        print("Error in deleting item with id '{0}'.".format(item_id))


def token_client_query(container, username):
    try:
        for item in container.query_items(
            query="SELECT * FROM my_container c WHERE c.username=@username",
            parameters=[{"name": "@username", "value": username}],
            partition_key=username,
        ):
            print(json.dumps(item, indent=True))
    except exceptions.CosmosHttpResponseError:
        print("Error in querying item(s)")


def run_sample():
    client = cosmos_client.CosmosClient(HOST, {"masterKey": MASTER_KEY})
#</configureConnectivity>


    try:
        try:
            db = client.create_database(DATABASE_ID)
        except exceptions.CosmosResourceExistsError:
            db = client.get_database_client(DATABASE_ID)

        try:
            container = db.create_container(
                id=CONTAINER_ID, partition_key=PARTITION_KEY
            )
        except exceptions.CosmosResourceExistsError:
            container = db.get_container_client(CONTAINER_ID)

        user = create_user_if_not_exists(db, USERNAME)

        # Permission to perform operations on all items inside a container
        permission_definition: Dict[str, Any] = {
            "id": CONTAINER_ALL_PERMISSION,
            "permissionMode": documents.PermissionMode.All,
            "resource": container.container_link,
        }

        permission = create_permission_if_not_exists(user, permission_definition)
        token = {}
        token[container.container_link] = permission.properties["_token"]

        # Use token to connect to database
        token_client = cosmos_client.CosmosClient(HOST, token)
        token_db = token_client.get_database_client(DATABASE_ID)
        token_container = token_db.get_container_client(CONTAINER_ID)

        ITEM_1_ID, ITEM_2_ID, ITEM_3_ID = "1", "2", "3"

        # Update or insert item if not exists
        token_client_upsert(token_container, USERNAME, ITEM_1_ID)
        token_client_upsert(token_container, USERNAME, ITEM_2_ID)
        token_client_upsert(token_container, USERNAME_2, ITEM_3_ID)

        # Read all items in the container, across all partitions
        token_client_read_all(token_container)

        # Read specific item
        token_client_read_item(token_container, USERNAME, ITEM_2_ID)

        # Query for items in a certain partition
        token_client_query(token_container, USERNAME_2)

        # Delete an item
        token_client_delete(token_container, USERNAME, ITEM_2_ID)

        # Give user read-only permission, for a specific partition
        user_2 = create_user_if_not_exists(db, USERNAME_2)
        permission_definition = {
            "id": PARTITION_READ_PERMISSION,
            "permissionMode": documents.PermissionMode.Read,
            "resource": container.container_link,
            "resourcePartitionKey": [USERNAME_2],
        }
        permission = create_permission_if_not_exists(user_2, permission_definition)
        read_token = {}
        read_token[container.container_link] = permission.properties["_token"]

        # Use token to connect to database
        token_client = cosmos_client.CosmosClient(HOST, read_token)
        token_db = token_client.get_database_client(DATABASE_ID)
        token_container = token_db.get_container_client(CONTAINER_ID)

        # Fails since this client has access to only items with partition key USERNAME_2 (ie. "user2")
        token_client_read_all(token_container)

        # Ok to read item(s) with partition key "user2"
        token_client_read_item(token_container, USERNAME_2, ITEM_3_ID)

        # Can't upsert or delete since it's read-only
        token_client_upsert(token_container, USERNAME_2, ITEM_3_ID)

        # Give user CRUD permissions, only for a specific item
        item_3 = token_container.read_item(item=ITEM_3_ID, partition_key=USERNAME_2)
        permission_list = list(user_2.list_permissions())
        for p in permission_list:
            user_2.delete_permission(p.get('id'))
        assert len(list(user_2.list_permissions())) == 0

        permission_definition = {
            "id": DOCUMENT_ALL_PERMISSION,
            "permissionMode": documents.PermissionMode.All,
            "resource": str(item_3.get('_self')) #this identifies the item with id "3"
        }

        permission = create_permission_if_not_exists(user_2, permission_definition)

        item_token = {}
        item_token[container.container_link] = permission.properties["_token"]

        # Use token to connect to database
        token_client = cosmos_client.CosmosClient(HOST, item_token)
        token_db = token_client.get_database_client(DATABASE_ID)
        token_container = token_db.get_container_client(CONTAINER_ID)

        # Fails since this client only has access to a specific item
        token_client_read_all(token_container)

        # Fails too, for same reason
        token_client_read_item(token_container, USERNAME, ITEM_1_ID)

        # Ok to perform operations on that specific item
        token_client_read_item(token_container, USERNAME_2, ITEM_3_ID)
        token_client_delete(token_container, USERNAME_2, ITEM_3_ID)

    except exceptions.CosmosHttpResponseError as e:
        print("\nrun_sample has caught an error. {0}".format(e.message))

    finally:
        print("\nrun_sample done")


if __name__ == "__main__":
    run_sample()



# The MIT License (MIT)
# Copyright (c) 2023 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import struct
from typing import NoReturn, Tuple, Union


class _UInt32:
    def __init__(self, value: int) -> None:
        self._value: int = value & 0xFFFFFFFF

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, new_value: int) -> None:
        self._value = new_value & 0xFFFFFFFF

    def __add__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        result = self.value + (other.value if isinstance(other, _UInt32) else other)
        return _UInt32(result & 0xFFFFFFFF)

    def __sub__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        result = self.value - (other.value if isinstance(other, _UInt32) else other)
        return _UInt32(result & 0xFFFFFFFF)

    def __mul__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        result = self.value * (other.value if isinstance(other, _UInt32) else other)
        return _UInt32(result & 0xFFFFFFFF)

    def __xor__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        result = self.value ^ (other.value if isinstance(other, _UInt32) else other)
        return _UInt32(result & 0xFFFFFFFF)

    def __lshift__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        result = self.value << (other.value if isinstance(other, _UInt32) else other)
        return _UInt32(result & 0xFFFFFFFF)

    def __ilshift__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        self._value = (self.value << (other.value if isinstance(other, _UInt32) else other)) & 0xFFFFFFFF
        return self

    def __rshift__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        result = self.value >> (other.value if isinstance(other, _UInt32) else other)
        return _UInt32(result & 0xFFFFFFFF)

    def __irshift__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        self._value = (self.value >> (other.value if isinstance(other, _UInt32) else other)) & 0xFFFFFFFF
        return self

    def __and__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        result = self.value & (other.value if isinstance(other, _UInt32) else other)
        return _UInt32(result & 0xFFFFFFFF)

    def __or__(self, other: Union[int, '_UInt32']) -> '_UInt32':
        if isinstance(other, _UInt32):
            return _UInt32(self.value | other.value)
        if isinstance(other, int):
            return _UInt32(self.value | other)
        raise TypeError("Unsupported type for OR operation")

    def __invert__(self) -> '_UInt32':
        return _UInt32(~self.value & 0xFFFFFFFF)

    def __eq__(self, other: Union[int, '_UInt32', object]) -> bool:
        return self.value == (other.value if isinstance(other, _UInt32) else other)

    def __ne__(self, other: Union[int, '_UInt32', object]) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other: Union[int, '_UInt32']) -> bool:
        return self.value < (other.value if isinstance(other, _UInt32) else other)

    def __gt__(self, other: Union[int, '_UInt32']) -> bool:
        return self.value > (other.value if isinstance(other, _UInt32) else other)

    def __le__(self, other: Union[int, '_UInt32']) -> bool:
        return self.value <= (other.value if isinstance(other, _UInt32) else other)

    def __ge__(self, other: Union[int, '_UInt32']) -> bool:
        return self.value >= (other.value if isinstance(other, _UInt32) else other)

    @staticmethod
    def encode_double_as_uint32(value: float) -> int:
        value_in_uint32 = struct.unpack('<I', struct.pack('<f', value))[0]
        mask = 0x80000000
        return (value_in_uint32 ^ mask) if value_in_uint32 < mask else (~value_in_uint32) + 1

    @staticmethod
    def decode_double_from_uint32(value: int) -> int:
        mask = 0x80000000
        value = ~(value - 1) if value < mask else value ^ mask
        return struct.unpack('<f', struct.pack('<I', value))[0]

    def __int__(self) -> int:
        return self.value

class _UInt64:
    def __init__(self, value: int) -> None:
        self._value: int = value & 0xFFFFFFFFFFFFFFFF

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, new_value: int) -> None:
        self._value = new_value & 0xFFFFFFFFFFFFFFFF

    def __add__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        result = self.value + (other.value if isinstance(other, _UInt64) else other)
        return _UInt64(result & 0xFFFFFFFFFFFFFFFF)

    def __sub__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        result = self.value - (other.value if isinstance(other, _UInt64) else other)
        return _UInt64(result & 0xFFFFFFFFFFFFFFFF)

    def __mul__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        result = self.value * (other.value if isinstance(other, _UInt64) else other)
        return _UInt64(result & 0xFFFFFFFFFFFFFFFF)

    def __xor__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        result = self.value ^ (other.value if isinstance(other, _UInt64) else other)
        return _UInt64(result & 0xFFFFFFFFFFFFFFFF)

    def __lshift__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        result = self.value << (other.value if isinstance(other, _UInt64) else other)
        return _UInt64(result & 0xFFFFFFFFFFFFFFFF)

    def __rshift__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        result = self.value >> (other.value if isinstance(other, _UInt64) else other)
        return _UInt64(result & 0xFFFFFFFFFFFFFFFF)

    def __and__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        result = self.value & (other.value if isinstance(other, _UInt64) else other)
        return _UInt64(result & 0xFFFFFFFFFFFFFFFF)

    def __or__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        if isinstance(other, _UInt64):
            return _UInt64(self.value | other.value)
        if isinstance(other, int):
            return _UInt64(self.value | other)
        raise TypeError("Unsupported type for OR operation")

    def __invert__(self) -> '_UInt64':
        return _UInt64(~self.value & 0xFFFFFFFFFFFFFFFF)

    def __eq__(self, other: Union[int, '_UInt64', object]) -> bool:
        return self.value == (other.value if isinstance(other, _UInt64) else other)

    def __ne__(self, other: Union[int, '_UInt64', object]) -> bool:
        return not self.__eq__(other)

    def __irshift__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        self._value = (self.value >> (other.value if isinstance(other, _UInt64) else other)) & 0xFFFFFFFFFFFFFFFF
        return self

    def __ilshift__(self, other: Union[int, '_UInt64']) -> '_UInt64':
        self._value = (self.value << (other.value if isinstance(other, _UInt64) else other)) & 0xFFFFFFFFFFFFFFFF
        return self

    def __lt__(self, other: Union[int, '_UInt64']) -> bool:
        return self.value < (other.value if isinstance(other, _UInt64) else other)

    def __gt__(self, other: Union[int, '_UInt64']) -> bool:
        return self.value > (other.value if isinstance(other, _UInt64) else other)

    def __le__(self, other: Union[int, '_UInt64']) -> bool:
        return self.value <= (other.value if isinstance(other, _UInt64) else other)

    def __ge__(self, other: Union[int, '_UInt64']) -> bool:
        return self.value >= (other.value if isinstance(other, _UInt64) else other)

    @staticmethod
    def encode_double_as_uint64(value: float) -> int:
        value_in_uint64 = struct.unpack('<Q', struct.pack('<d', value))[0]
        mask = 0x8000000000000000
        return (value_in_uint64 ^ mask) if value_in_uint64 < mask else (~value_in_uint64) + 1

    @staticmethod
    def decode_double_from_uint64(value: int) -> int:
        mask = 0x8000000000000000
        value = ~(value - 1) if value < mask else value ^ mask
        return struct.unpack('<d', struct.pack('<Q', value))[0]

    def __int__(self) -> int:
        return self.value


class _UInt128:
    def __init__(self, low: Union[int, _UInt64], high: Union[int, _UInt64]) -> None:
        if isinstance(low, _UInt64):
            self.low = low
        else:
            self.low = _UInt64(low)
        if isinstance(high, _UInt64):
            self.high = high
        else:
            self.high = _UInt64(high)

    def __add__(self, other: '_UInt128') -> '_UInt128':
        low = self.low + other.low
        high = self.high + other.high + _UInt64(int(low.value > 0xFFFFFFFFFFFFFFFF))
        return _UInt128(low & 0xFFFFFFFFFFFFFFFF, high & 0xFFFFFFFFFFFFFFFF)

    def __sub__(self, other: '_UInt128') -> '_UInt128':
        borrow = _UInt64(0)
        if self.low.value < other.low.value:
            borrow = _UInt64(1)

        low = (self.low - other.low) & 0xFFFFFFFFFFFFFFFF
        high = (self.high - other.high - borrow) & 0xFFFFFFFFFFFFFFFF
        return _UInt128(low, high)

    def __mul__(self, other: '_UInt128') -> NoReturn:
        # Multiplication logic here for 128 bits
        raise NotImplementedError()

    def __xor__(self, other: '_UInt128') -> '_UInt128':
        low = self.low ^ other.low
        high = self.high ^ other.high
        return _UInt128(low, high)

    def __and__(self, other: '_UInt128') -> '_UInt128':
        low = self.low & other.low
        high = self.high & other.high
        return _UInt128(low, high)

    def __or__(self, other: '_UInt128') -> '_UInt128':
        low = self.low | other.low
        high = self.high | other.high
        return _UInt128(low, high)

    def __lshift__(self, shift: '_UInt128') -> NoReturn:
        # Left shift logic for 128 bits
        raise NotImplementedError()

    def __rshift__(self, shift: '_UInt128') -> NoReturn:
        # Right shift logic for 128 bits
        raise NotImplementedError()

    def get_low(self) -> _UInt64:
        return self.low

    def get_high(self) -> _UInt64:
        return self.high

    def as_tuple(self) -> Tuple[int, int]:
        return self.low.value, self.high.value

    def as_hex(self) -> str:
        return hex(self.high.value)[2:].zfill(16) + hex(self.low.value)[2:].zfill(16)

    def as_int(self) -> int:
        return (self.high.value << 64) | self.low.value

    def __str__(self) -> str:
        return str(self.as_int())

    def to_byte_array(self) -> bytearray:
        high_bytes = self.high.value.to_bytes(8, byteorder='little')
        low_bytes = self.low.value.to_bytes(8, byteorder='little')
        byte_array = bytearray(low_bytes + high_bytes)
        return byte_array



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest

import azure.cosmos.exceptions as exceptions
import test_config
from azure.cosmos import CosmosClient, cosmos_client
from azure.cosmos import ThroughputProperties, PartitionKey

@pytest.mark.cosmosLong
class TestAutoScale(unittest.TestCase):
    client: CosmosClient = None
    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    connectionPolicy = test_config.TestConfig.connectionPolicy

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

        cls.client = cosmos_client.CosmosClient(cls.host, cls.masterKey)
        cls.created_database = cls.client.get_database_client(test_config.TestConfig.TEST_DATABASE_ID)

    def test_autoscale_create_container(self):
        container_id = None
        try:
            container_id = 'auto_scale'
            created_container = self.created_database.create_container(
                id=container_id,
                partition_key=PartitionKey(path="/id"),
                offer_throughput=ThroughputProperties(auto_scale_max_throughput=7000, auto_scale_increment_percent=0)

            )
            created_container_properties = created_container.get_throughput()
            # Testing the input value of the max_throughput
            assert created_container_properties.auto_scale_max_throughput == 7000
            assert created_container_properties.auto_scale_increment_percent == 0
            assert created_container_properties.offer_throughput is None

            self.created_database.delete_container(created_container)

            # Testing the incorrect passing of an input value of the max_throughput to verify negative behavior
            with self.assertRaises(exceptions.CosmosHttpResponseError) as e:
                self.created_database.create_container(
                    id='container_with_wrong_auto_scale_settings',
                    partition_key=PartitionKey(path="/id"),
                    offer_throughput=ThroughputProperties(auto_scale_max_throughput=-200, auto_scale_increment_percent=0))
            assert "Requested throughput -200 is less than required minimum throughput 1000" in str(e.exception)

            # Testing auto_scale_settings for the create_container_if_not_exists method
            container_id = 'auto_scale_2'
            created_container = self.created_database.create_container_if_not_exists(
                id=container_id,
                partition_key=PartitionKey(path="/id"),
                offer_throughput=ThroughputProperties(auto_scale_max_throughput=1000, auto_scale_increment_percent=3)
            )
            created_container_properties = created_container.get_throughput()
            # Testing the input value of the max_throughput
            assert created_container_properties.auto_scale_max_throughput == 1000
            # Testing the input value of the increment_percentage
            assert created_container_properties.auto_scale_increment_percent == 3
        finally:
            self.created_database.delete_container(container_id)

    def test_autoscale_create_database(self):
        database_id = "db_auto_scale_" + str(uuid.uuid4())
        try:
            # Testing auto_scale_settings for the create_database method
            created_database = self.client.create_database(database_id, offer_throughput=ThroughputProperties(
                auto_scale_max_throughput=5000,
                auto_scale_increment_percent=2))
            created_db_properties = created_database.get_throughput()
            # Testing the input value of the max_throughput
            assert created_db_properties.auto_scale_max_throughput == 5000
            # Testing the input value of the increment_percentage
            assert created_db_properties.auto_scale_increment_percent == 2

            self.client.delete_database(created_database.id)

            # Testing auto_scale_settings for the create_database_if_not_exists method
            database_id = "db_auto_scale_2_" + str(uuid.uuid4())
            created_database = self.client.create_database_if_not_exists(database_id,
                                                                         offer_throughput=ThroughputProperties(
                                                                             auto_scale_max_throughput=9000,
                                                                             auto_scale_increment_percent=11))
            created_db_properties = created_database.get_throughput()
            # Testing the input value of the max_throughput
            assert created_db_properties.auto_scale_max_throughput == 9000
            # Testing the input value of the increment_percentage
            assert created_db_properties.auto_scale_increment_percent == 11
        finally:
            self.client.delete_database(database_id)

    def test_autoscale_replace_throughput(self):
        database_id = "replace_db" + str(uuid.uuid4())
        container_id = None
        try:
            created_database = self.client.create_database(database_id, offer_throughput=ThroughputProperties(
                auto_scale_max_throughput=5000,
                auto_scale_increment_percent=2))
            created_database.replace_throughput(
                throughput=ThroughputProperties(auto_scale_max_throughput=7000, auto_scale_increment_percent=20))
            created_db_properties = created_database.get_throughput()
            # Testing the input value of the max_throughput
            assert created_db_properties.auto_scale_max_throughput == 7000
            # Testing the input value of the increment_percentage
            assert created_db_properties.auto_scale_increment_percent == 20
            self.client.delete_database(database_id)

            container_id = "container_with_auto_scale_settings" + str(uuid.uuid4())
            created_container = self.created_database.create_container(
                id=container_id,
                partition_key=PartitionKey(path="/id"),
                offer_throughput=ThroughputProperties(auto_scale_max_throughput=5000, auto_scale_increment_percent=0))
            created_container.replace_throughput(
                throughput=ThroughputProperties(auto_scale_max_throughput=7000, auto_scale_increment_percent=20))
            created_container_properties = created_container.get_throughput()
            # Testing the input value of the replaced auto_scale settings
            assert created_container_properties.auto_scale_max_throughput == 7000
            assert created_container_properties.auto_scale_increment_percent == 20

        finally:
            self.created_database.delete_container(container_id)


if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import random
import time
import unittest
import os

import pytest

import azure.cosmos.cosmos_client as cosmos_client
import test_config
from azure.cosmos import DatabaseProxy, PartitionKey, ContainerProxy
from azure.cosmos.exceptions import CosmosHttpResponseError


def run_queries(container, iterations):
    ret_list = []
    for i in range(iterations):
        curr = str(random.randint(0, 10))
        query = 'SELECT * FROM c WHERE c.attr1=' + curr + ' order by c.attr1'
        qlist = list(container.query_items(query=query, enable_cross_partition_query=True))
        ret_list.append((curr, qlist))
    for ret in ret_list:
        curr = ret[0]
        if len(ret[1]) != 0:
            for results in ret[1]:
                attr_number = results['attr1']
                assert str(attr_number) == curr  # verify that all results match their randomly generated attributes
        print("validation succeeded for all query results")


@pytest.mark.cosmosSplit
class TestPartitionSplitQuery(unittest.TestCase):
    database: DatabaseProxy = None
    container: ContainerProxy = None
    client: cosmos_client.CosmosClient = None
    configs = test_config.TestConfig
    host = configs.host
    masterKey = configs.masterKey
    throughput = 400
    TEST_DATABASE_ID = configs.TEST_DATABASE_ID
    TEST_CONTAINER_ID = "Single-partition-container-without-throughput"
    MAX_TIME = 60 * 10  # 10 minutes for the test to complete, should be enough for partition split to complete

    @classmethod
    def setUpClass(cls):
        cls.client = cosmos_client.CosmosClient(cls.host, cls.masterKey)
        cls.database = cls.client.get_database_client(cls.TEST_DATABASE_ID)
        cls.container = cls.database.create_container(
            id=cls.TEST_CONTAINER_ID,
            partition_key=PartitionKey(path="/id"),
            offer_throughput=cls.throughput)
        if cls.host == "https://localhost:8081/":
            os.environ["AZURE_COSMOS_DISABLE_NON_STREAMING_ORDER_BY"] = "True"

    @classmethod
    def tearDownClass(cls) -> None:
        try:
            cls.database.delete_container(cls.container.id)
        except CosmosHttpResponseError:
            pass

    def test_partition_split_query(self):
        for i in range(100):
            body = test_config.get_test_item()
            self.container.create_item(body=body)

        start_time = time.time()
        print("created items, changing offer to 11k and starting queries")
        self.container.replace_throughput(11000)
        offer_time = time.time()
        print("changed offer to 11k")
        print("--------------------------------")
        print("now starting queries")

        run_queries(self.container, 100)  # initial check for queries before partition split
        print("initial check succeeded, now reading offer until replacing is done")
        offer = self.container.get_throughput()
        while True:
            if time.time() - start_time > self.MAX_TIME:  # timeout test at 10 minutes
                self.skipTest("Partition split didn't complete in time")
            if offer.properties['content'].get('isOfferReplacePending', False):
                time.sleep(30)  # wait for the offer to be replaced, check every 30 seconds
                offer = self.container.get_throughput()
            else:
                print("offer replaced successfully, took around {} seconds".format(time.time() - offer_time))
                run_queries(self.container, 100)  # check queries work post partition split
                self.assertTrue(offer.offer_throughput > self.throughput)
                return


if __name__ == "__main__":
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import os
import unittest
import uuid

import pytest

import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.exceptions as exceptions
import test_config
from azure.cosmos import http_constants, DatabaseProxy
from azure.cosmos._execution_context.base_execution_context import _QueryExecutionContextBase
from azure.cosmos._execution_context.query_execution_info import _PartitionedQueryExecutionInfo
from azure.cosmos.documents import _DistinctType
from azure.cosmos.partition_key import PartitionKey

@pytest.mark.cosmosCircuitBreaker
@pytest.mark.cosmosQuery
class TestCrossPartitionQuery(unittest.TestCase):
    """Test to ensure escaping of non-ascii characters from partition key"""

    created_db: DatabaseProxy = None
    client: cosmos_client.CosmosClient = None
    config = test_config.TestConfig
    host = config.host
    masterKey = config.masterKey
    connectionPolicy = config.connectionPolicy
    TEST_DATABASE_ID = config.TEST_DATABASE_ID
    TEST_CONTAINER_ID = "Multi Partition Test Collection With Custom PK " + str(uuid.uuid4())

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

        use_multiple_write_locations = False
        if os.environ.get("AZURE_COSMOS_ENABLE_CIRCUIT_BREAKER", "False") == "True":
            use_multiple_write_locations = True
        cls.client = cosmos_client.CosmosClient(cls.host, cls.masterKey, multiple_write_locations=use_multiple_write_locations)
        cls.created_db = cls.client.get_database_client(cls.TEST_DATABASE_ID)
        if cls.host == "https://localhost:8081/":
            os.environ["AZURE_COSMOS_DISABLE_NON_STREAMING_ORDER_BY"] = "True"

    def setUp(self):
        self.created_container = self.created_db.create_container(
            id=self.TEST_CONTAINER_ID,
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_5_PARTITIONS)

    def tearDown(self):
        try:
            self.created_db.delete_container(self.TEST_CONTAINER_ID)
        except exceptions.CosmosHttpResponseError:
            pass

    def test_first_and_last_slashes_trimmed_for_query_string(self):
        doc_id = 'myId' + str(uuid.uuid4())
        document_definition = {'pk': 'pk', 'id': doc_id}
        self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            partition_key='pk'
        )
        iter_list = list(query_iterable)
        self.assertEqual(iter_list[0]['id'], doc_id)

    def test_query_change_feed_with_pk(self):
        # The test targets partition #3
        partition_key = "pk"

        # Read change feed without passing any options
        query_iterable = self.created_container.query_items_change_feed()
        iter_list = list(query_iterable)
        self.assertEqual(len(iter_list), 0)

        # Read change feed from current should return an empty list
        query_iterable = self.created_container.query_items_change_feed(partition_key=partition_key)
        iter_list = list(query_iterable)
        self.assertEqual(len(iter_list), 0)
        self.assertTrue('etag' in self.created_container.client_connection.last_response_headers)
        self.assertNotEqual(self.created_container.client_connection.last_response_headers['etag'], '')

        # Read change feed from beginning should return an empty list
        query_iterable = self.created_container.query_items_change_feed(
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        iter_list = list(query_iterable)
        self.assertEqual(len(iter_list), 0)
        self.assertTrue('etag' in self.created_container.client_connection.last_response_headers)
        continuation1 = self.created_container.client_connection.last_response_headers['etag']
        self.assertNotEqual(continuation1, '')

        # Create a document. Read change feed should return be able to read that document
        document_definition = {'pk': 'pk', 'id': 'doc1'}
        self.created_container.create_item(body=document_definition)
        query_iterable = self.created_container.query_items_change_feed(
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        iter_list = list(query_iterable)
        self.assertEqual(len(iter_list), 1)
        self.assertEqual(iter_list[0]['id'], 'doc1')
        self.assertTrue('etag' in self.created_container.client_connection.last_response_headers)
        continuation2 = self.created_container.client_connection.last_response_headers['etag']
        self.assertNotEqual(continuation2, '')
        self.assertNotEqual(continuation2, continuation1)

        # Create two new documents. Verify that change feed contains the 2 new documents
        # with page size 1 and page size 100
        document_definition = {'pk': 'pk', 'id': 'doc2'}
        self.created_container.create_item(body=document_definition)
        document_definition = {'pk': 'pk', 'id': 'doc3'}
        self.created_container.create_item(body=document_definition)

        for pageSize in [1, 100]:
            # verify iterator
            query_iterable = self.created_container.query_items_change_feed(
                continuation=continuation2,
                max_item_count=pageSize,
                partition_key=partition_key
            )
            it = query_iterable.__iter__()
            expected_ids = 'doc2.doc3.'
            actual_ids = ''
            for item in it:
                actual_ids += item['id'] + '.'
            self.assertEqual(actual_ids, expected_ids)

            # verify by_page
            # the options is not copied, therefore it need to be restored
            query_iterable = self.created_container.query_items_change_feed(
                continuation=continuation2,
                max_item_count=pageSize,
                partition_key=partition_key
            )
            count = 0
            expected_count = 2
            all_fetched_res = []
            for page in query_iterable.by_page():
                fetched_res = list(page)
                self.assertEqual(len(fetched_res), min(pageSize, expected_count - count))
                count += len(fetched_res)
                all_fetched_res.extend(fetched_res)

            actual_ids = ''
            for item in all_fetched_res:
                actual_ids += item['id'] + '.'
            self.assertEqual(actual_ids, expected_ids)

        # verify reading change feed from the beginning
        query_iterable = self.created_container.query_items_change_feed(
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        expected_ids = ['doc1', 'doc2', 'doc3']
        it = query_iterable.__iter__()
        for i in range(0, len(expected_ids)):
            doc = next(it)
            self.assertEqual(doc['id'], expected_ids[i])
        self.assertTrue('etag' in self.created_container.client_connection.last_response_headers)
        continuation3 = self.created_container.client_connection.last_response_headers['etag']

        # verify reading empty change feed
        query_iterable = self.created_container.query_items_change_feed(
            continuation=continuation3,
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        iter_list = list(query_iterable)
        self.assertEqual(len(iter_list), 0)

    def test_populate_query_metrics(self):
        doc_id = 'MyId' + str(uuid.uuid4())
        document_definition = {'pk': 'pk', 'id': doc_id}
        self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            partition_key='pk',
            populate_query_metrics=True
        )

        iter_list = list(query_iterable)
        self.assertEqual(iter_list[0]['id'], doc_id)

        metrics_header_name = 'x-ms-documentdb-query-metrics'
        self.assertTrue(metrics_header_name in self.created_container.client_connection.last_response_headers)
        metrics_header = self.created_container.client_connection.last_response_headers[metrics_header_name]
        # Validate header is well-formed: "key1=value1;key2=value2;etc"
        metrics = metrics_header.split(';')
        self.assertTrue(len(metrics) > 1)
        self.assertTrue(all(['=' in x for x in metrics]))

    def test_populate_index_metrics(self):
        doc_id = 'MyId' + str(uuid.uuid4())
        document_definition = {'pk': 'pk', 'id': doc_id}
        self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            partition_key='pk',
            populate_index_metrics=True
        )

        iter_list = list(query_iterable)
        self.assertEqual(iter_list[0]['id'], doc_id)

        INDEX_HEADER_NAME = http_constants.HttpHeaders.IndexUtilization
        self.assertTrue(INDEX_HEADER_NAME in self.created_container.client_connection.last_response_headers)
        index_metrics = self.created_container.client_connection.last_response_headers[INDEX_HEADER_NAME]
        self.assertIsNotNone(index_metrics)
        expected_index_metrics = {'UtilizedSingleIndexes': [{'FilterExpression': '', 'IndexSpec': '/pk/?',
                                                             'FilterPreciseSet': True, 'IndexPreciseSet': True,
                                                             'IndexImpactScore': 'High'}],
                                  'PotentialSingleIndexes': [], 'UtilizedCompositeIndexes': [],
                                  'PotentialCompositeIndexes': []}
        self.assertDictEqual(expected_index_metrics, index_metrics)

    def test_get_query_plan_through_gateway(self):
        self._validate_query_plan(query="Select top 10 value count(c.id) from c",
                                  container_link=self.created_container.container_link,
                                  top=10,
                                  order_by=[],
                                  aggregate=['Count'],
                                  select_value=True,
                                  offset=None,
                                  limit=None,
                                  distinct=_DistinctType.NoneType)

        self._validate_query_plan(query="Select * from c order by c._ts offset 5 limit 10",
                                  container_link=self.created_container.container_link,
                                  top=None,
                                  order_by=['Ascending'],
                                  aggregate=[],
                                  select_value=False,
                                  offset=5,
                                  limit=10,
                                  distinct=_DistinctType.NoneType)

        self._validate_query_plan(query="Select distinct value c.id from c order by c.id",
                                  container_link=self.created_container.container_link,
                                  top=None,
                                  order_by=['Ascending'],
                                  aggregate=[],
                                  select_value=True,
                                  offset=None,
                                  limit=None,
                                  distinct=_DistinctType.Ordered)

    def _validate_query_plan(self, query, container_link, top, order_by, aggregate, select_value, offset, limit,
                             distinct):
        query_plan_dict = self.client.client_connection._GetQueryPlanThroughGateway(query, container_link)
        query_execution_info = _PartitionedQueryExecutionInfo(query_plan_dict)
        self.assertTrue(query_execution_info.has_rewritten_query())
        self.assertEqual(query_execution_info.has_distinct_type(), distinct != "None")
        self.assertEqual(query_execution_info.get_distinct_type(), distinct)
        self.assertEqual(query_execution_info.has_top(), top is not None)
        self.assertEqual(query_execution_info.get_top(), top)
        self.assertEqual(query_execution_info.has_order_by(), len(order_by) > 0)
        self.assertListEqual(query_execution_info.get_order_by(), order_by)
        self.assertEqual(query_execution_info.has_aggregates(), len(aggregate) > 0)
        self.assertListEqual(query_execution_info.get_aggregates(), aggregate)
        self.assertEqual(query_execution_info.has_select_value(), select_value)
        self.assertEqual(query_execution_info.has_offset(), offset is not None)
        self.assertEqual(query_execution_info.get_offset(), offset)
        self.assertEqual(query_execution_info.has_limit(), limit is not None)
        self.assertEqual(query_execution_info.get_limit(), limit)

    def test_unsupported_queries(self):
        queries = ['SELECT COUNT(1) FROM c', 'SELECT COUNT(1) + 5 FROM c', 'SELECT COUNT(1) + SUM(c) FROM c']
        for query in queries:
            query_iterable = self.created_container.query_items(query=query, enable_cross_partition_query=True)
            try:
                list(query_iterable)
                self.fail()
            except exceptions.CosmosHttpResponseError as e:
                self.assertEqual(e.status_code, 400)

    def test_query_with_non_overlapping_pk_ranges(self):
        query_iterable = self.created_container.query_items("select * from c where c.pk='1' or c.pk='2'",
                                                            enable_cross_partition_query=True)
        self.assertListEqual(list(query_iterable), [])

    def test_offset_limit(self):
        values = []
        for i in range(10):
            document_definition = {'pk': i, 'id': 'myId' + str(uuid.uuid4()), 'value': i // 3}
            values.append(self.created_container.create_item(body=document_definition)['pk'])

        self.config._validate_distinct_offset_limit(
            created_collection=self.created_container,
            query='SELECT DISTINCT c["value"] from c ORDER BY c.pk OFFSET 0 LIMIT 2',
            results=[0, 1])

        self.config._validate_distinct_offset_limit(
            created_collection=self.created_container,
            query='SELECT DISTINCT c["value"] from c ORDER BY c.pk OFFSET 2 LIMIT 2',
            results=[2, 3])

        self.config._validate_distinct_offset_limit(
            created_collection=self.created_container,
            query='SELECT DISTINCT c["value"] from c ORDER BY c.pk OFFSET 4 LIMIT 3',
            results=[])

        self.config._validate_offset_limit(created_collection=self.created_container,
                                           query='SELECT * from c ORDER BY c.pk OFFSET 0 LIMIT 5',
                                           results=values[:5])

        self.config._validate_offset_limit(created_collection=self.created_container,
                                           query='SELECT * from c ORDER BY c.pk OFFSET 5 LIMIT 10',
                                           results=values[5:])

        self.config._validate_offset_limit(created_collection=self.created_container,
                                           query='SELECT * from c ORDER BY c.pk OFFSET 10 LIMIT 5',
                                           results=[])

        self.config._validate_offset_limit(created_collection=self.created_container,
                                           query='SELECT * from c ORDER BY c.pk OFFSET 100 LIMIT 1',
                                           results=[])

    def test_distinct_on_different_types_and_field_orders(self):
        self.payloads = [
            {'f1': 1, 'f2': 'value', 'f3': 100000000000000000, 'f4': [1, 2, '3'], 'f5': {'f6': {'f7': 2}}},
            {'f2': '\'value', 'f4': [1.0, 2, '3'], 'f5': {'f6': {'f7': 2.0}}, 'f1': 1.0, 'f3': 100000000000000000.00},
            {'f3': 100000000000000000.0, 'f5': {'f6': {'f7': 2}}, 'f2': '\'value', 'f1': 1, 'f4': [1, 2.0, '3']}
        ]
        self.OriginalExecuteFunction = _QueryExecutionContextBase.__next__
        _QueryExecutionContextBase.__next__ = self._MockNextFunction

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f1 from c",
            expected_results=[1],
            get_mock_result=lambda x, i: (None, x[i]["f1"])
        )

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f2 from c",
            expected_results=['value', '\'value'],
            get_mock_result=lambda x, i: (None, x[i]["f2"])
        )

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f2 from c order by c.f2",
            expected_results=['\'value', 'value'],
            get_mock_result=lambda x, i: (x[i]["f2"], x[i]["f2"])
        )

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f3 from c",
            expected_results=[100000000000000000],
            get_mock_result=lambda x, i: (None, x[i]["f3"])
        )

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f4 from c",
            expected_results=[[1, 2, '3']],
            get_mock_result=lambda x, i: (None, x[i]["f4"])
        )

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f5.f6 from c",
            expected_results=[{'f7': 2}],
            get_mock_result=lambda x, i: (None, x[i]["f5"]["f6"])
        )

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct c.f1, c.f2, c.f3 from c",
            expected_results=[self.payloads[0], self.payloads[1]],
            get_mock_result=lambda x, i: (None, x[i])
        )

        self._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct c.f1, c.f2, c.f3 from c order by c.f1",
            expected_results=[self.payloads[0], self.payloads[1]],
            get_mock_result=lambda x, i: (i, x[i])
        )

        _QueryExecutionContextBase.__next__ = self.OriginalExecuteFunction
        _QueryExecutionContextBase.next = self.OriginalExecuteFunction

    def test_paging_with_continuation_token(self):
        document_definition = {'pk': 'pk', 'id': '1'}
        self.created_container.create_item(body=document_definition)
        document_definition = {'pk': 'pk', 'id': '2'}
        self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            partition_key='pk',
            max_item_count=1
        )
        pager = query_iterable.by_page()
        pager.next()
        token = pager.continuation_token
        second_page = list(pager.next())[0]

        pager = query_iterable.by_page(token)
        second_page_fetched_with_continuation_token = list(pager.next())[0]

        self.assertEqual(second_page['id'], second_page_fetched_with_continuation_token['id'])

    def test_cross_partition_query_with_continuation_token(self):
        document_definition = {'pk': 'pk1', 'id': str(uuid.uuid4())}
        self.created_container.create_item(body=document_definition)
        document_definition = {'pk': 'pk2', 'id': str(uuid.uuid4())}
        self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            enable_cross_partition_query=True,
            max_item_count=1,
        )
        pager = query_iterable.by_page()
        pager.next()
        token = pager.continuation_token
        second_page = list(pager.next())[0]

        pager = query_iterable.by_page(token)
        second_page_fetched_with_continuation_token = list(pager.next())[0]

        self.assertEqual(second_page['id'], second_page_fetched_with_continuation_token['id'])

    def _validate_distinct_on_different_types_and_field_orders(self, collection, query, expected_results,
                                                               get_mock_result):
        self.count = 0
        self.get_mock_result = get_mock_result
        query_iterable = collection.query_items(query, enable_cross_partition_query=True)
        results = list(query_iterable)
        for i in range(len(expected_results)):
            if isinstance(results[i], dict):
                self.assertDictEqual(results[i], expected_results[i])
            elif isinstance(results[i], list):
                self.assertListEqual(results[i], expected_results[i])
            else:
                self.assertEqual(results[i], expected_results[i])
        self.count = 0

    def test_value_max_query(self):
        query = "Select value max(c.version) FROM c where c.isComplete = true and c.lookupVersion = @lookupVersion"
        query_results = self.created_container.query_items(query, parameters=[
            {"name": "@lookupVersion", "value": "console_csat"}  # cspell:disable-line
        ], enable_cross_partition_query=True)

        self.assertListEqual(list(query_results), [None])

    def test_continuation_token_size_limit_query(self):
        for i in range(1, 1000):
            self.created_container.create_item(body=dict(pk='123', id=str(i), some_value=str(i % 3)))
        query = "Select * from c where c.some_value='2'"
        response_query = self.created_container.query_items(query, partition_key='123', max_item_count=100,
                                                            continuation_token_limit=1)
        pager = response_query.by_page()
        pager.next()
        token = pager.continuation_token
        # Continuation token size should be below 1kb
        self.assertLessEqual(len(token.encode('utf-8')), 1024)
        pager.next()
        token = pager.continuation_token

        # verify a second time
        self.assertLessEqual(len(token.encode('utf-8')), 1024)

    def test_cross_partition_query_response_hook(self):
        created_collection = self.created_db.create_container_if_not_exists(
            id="query_response_hook_test" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=12000
        )
        items = [
            {'id': str(uuid.uuid4()), 'pk': '0', 'val': 5},
            {'id': str(uuid.uuid4()), 'pk': '1', 'val': 10},
            {'id': str(uuid.uuid4()), 'pk': '0', 'val': 5},
            {'id': str(uuid.uuid4()), 'pk': '1', 'val': 10},
            {'id': str(uuid.uuid4()), 'pk': '0', 'val': 5},
            {'id': str(uuid.uuid4()), 'pk': '1', 'val': 10}
        ]

        for item in items:
            created_collection.create_item(body=item)

        response_hook = test_config.ResponseHookCaller()
        item_list = [item for item in created_collection.query_items("select * from c", enable_cross_partition_query=True, response_hook=response_hook)]
        assert len(item_list) == 6
        assert response_hook.count == 2
        self.created_db.delete_container(created_collection.id)

    def test_cross_partition_query_pagination_with_max_item_count(self):
        """Test cross-partition pagination showing per-page limits and total results."""
        created_collection = self.created_db.create_container(
            "cross_partition_pagination_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_5_PARTITIONS)
        
        # Create 30 items across 3 different partitions
        total_items = 30
        items_per_partition = 10
        partitions = ["pk_1", "pk_2", "pk_3"]
        
        for pk in partitions:
            for i in range(items_per_partition):
                document_definition = {
                    'pk': pk,
                    'id': f'{pk}_item_{i}',
                    'value': i
                }
                created_collection.create_item(body=document_definition)
        
        # Test cross-partition query with max_item_count
        max_items_per_page = 8
        query = "SELECT * FROM c ORDER BY c['value']"
        query_iterable = created_collection.query_items(
            query=query,
            enable_cross_partition_query=True,
            max_item_count=max_items_per_page
        )
        
        # Iterate through pages and verify per-page counts
        all_fetched_results = []
        page_count = 0
        item_pages = query_iterable.by_page()
        
        for page in item_pages:
            page_count += 1
            items_in_page = list(page)
            all_fetched_results.extend(items_in_page)
            
            # Each page should have at most max_item_count items
            # In cross-partition queries, pages may have fewer items
            self.assertLessEqual(len(items_in_page), max_items_per_page)
            self.assertGreater(len(items_in_page), 0)  # Pages should not be empty
        
        # Verify total results match expected count
        self.assertEqual(len(all_fetched_results), total_items)
        
        # Verify we got multiple pages
        self.assertGreater(page_count, 1)
        
        self.created_db.delete_container(created_collection.id)
    
    def test_cross_partition_query_pagination_counting_results(self):
        """Test counting total results while paginating across partitions."""
        created_collection = self.created_db.create_container(
            "cross_partition_count_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_5_PARTITIONS)
        
        # Create items across multiple partitions with different counts
        partitions_config = [
            ("partition_a", 5),
            ("partition_b", 8),
            ("partition_c", 3),
            ("partition_d", 12)
        ]
        
        total_expected = 0
        for pk, count in partitions_config:
            for i in range(count):
                document_definition = {
                    'pk': pk,
                    'id': f'{pk}_item_{i}',
                    'name': f'Item {i} in {pk}'
                }
                created_collection.create_item(body=document_definition)
                total_expected += 1
        
        # Query across partitions with pagination
        max_items_per_page = 5
        query = "SELECT * FROM c"
        query_iterable = created_collection.query_items(
            query=query,
            enable_cross_partition_query=True,
            max_item_count=max_items_per_page
        )
        
        # Count items across all pages
        total_count = 0
        page_count = 0
        page_sizes = []
        
        item_pages = query_iterable.by_page()
        for page in item_pages:
            page_count += 1
            items = list(page)
            page_size = len(items)
            page_sizes.append(page_size)
            total_count += page_size
            
            # Verify page size constraints
            self.assertLessEqual(page_size, max_items_per_page)
            self.assertGreater(page_size, 0)
        
        # Verify total count matches what we inserted
        self.assertEqual(total_count, total_expected)
        
        # Verify we processed multiple pages
        self.assertGreater(page_count, 1)
        
        self.created_db.delete_container(created_collection.id)

    def _MockNextFunction(self):
        if self.count < len(self.payloads):
            item, result = self.get_mock_result(self.payloads, self.count)
            self.count += 1
            if item is not None:
                return {'orderByItems': [{'item': item}], '_rid': 'fake_rid', 'payload': result}
            else:
                return result
        else:
            raise StopIteration


if __name__ == "__main__":
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import uuid

from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
from azure.cosmos.partition_key import PartitionKey
from azure.cosmos import ThroughputProperties

import asyncio
import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://azure.microsoft.com/documentation/articles/documentdb-create-account/
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates the basic CRUD operations for Control Plane responses for Azure Cosmos
#
# 1. Create Database
#    1.1 - Basic Create Database
#    1.2 - Create Database if not previously existing
#    1.3 - Create Database if not previously existing
#    1.4 - Database Read
#    1.5 - Database Replace Throughput
#
# 2. Create Container
#    2.1 - Basic Create Container
#    2.2 - Create container if not previously existing
#    2.3 - Create container if not previously existing
#    2.4 - Container Read
#    2.5 - Container Replace Throughput
#
# ----------------------------------------------------------------------------------------------------------
# Note -
#
# Running this sample will create (and delete) multiple Containers on your account.
# Each time a Container is created the account will be billed for 1 hour of usage based on
# the provisioned throughput (RU/s) of that account.
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id'] + str(uuid.uuid4())
CONTAINER_ID = config.settings['container_id'] + str(uuid.uuid4())


async def create_db(client, id):
    # create database with return properties parameter set to true
    print("\n1.1 Create Database")

    try:
        response_properties = await client.create_database(id=id, return_properties=True)
        print('Database with id \'{0}\' created'.format(id))
        print(response_properties[1].get_response_headers())
    except exceptions.CosmosResourceExistsError:
        print('A database with id \'{0}\' already exists'.format(id))


async def create_db_if_not_exists(client, id):
    print("\n1.2 Create Database if not exists")

    try:
        response_properties = await client.create_database_if_not_exists(id=id, return_properties=True)
        print('Database with id \'{0}\' created or retrieved'.format(id))
        print(response_properties[1].get_response_headers())
    except exceptions.CosmosHttpResponseError as e:
        print('Error creating database: {0}'.format(e.message))


async def create_db_if_exists(client, id):
    print("\n1.3 Get Database if exists")

    try:
        response_properties = await client.create_database_if_not_exists(id=id, return_properties=True)
        print('Database with id \'{0}\' found'.format(id))
        print(response_properties[1].get_response_headers())
    except exceptions.CosmosResourceNotFoundError:
        print('Database with id \'{0}\' does not exist'.format(id))


async def read_db(client, id):
    print("\n1.4 Read Database")

    try:
        database = client.get_database_client(id)
        properties = await database.read()
        print('Database with id \'{0}\' was found'.format(id))
    except exceptions.CosmosResourceNotFoundError:
        print('A database with id \'{0}\' does not exist'.format(id))


async def replace_throughput_db(client, id):
    print("\n1.5 Replace Database Throughput")

    try:
        database = await client.create_database(id=id+"1", offer_throughput=400)
        replace_throughput_value = 500
        properties = await database.replace_throughput(replace_throughput_value)
        print('Database throughput changed to 800 RU/s')
        print(properties.get_response_headers())
    except exceptions.CosmosResourceNotFoundError:
        print('Database with id \'{0}\' does not exist'.format(id))
    except exceptions.CosmosHttpResponseError as e:
        print('Error changing database throughput: {0}'.format(e.message))


async def create_container(db, id):
    """ Execute basic container creation.
    This will create containers with 400 RUs with different indexing, partitioning, and storage options """

    partition_key = PartitionKey(path='/id', kind='Hash')
    print("\n2.1 Create Container - Basic")

    try:
        response_properties = await db.create_container(id=id, partition_key=partition_key, return_properties=True)
        print('Container with id \'{0}\' created'.format(id))
        print(response_properties[1].get_response_headers())
    except exceptions.CosmosResourceExistsError:
        print('A container with id \'{0}\' already exists'.format(id))


async def create_container_if_not_exists(db, id):
    print("\n2.2 Create Container if not exists")

    partition_key = PartitionKey(path='/id', kind='Hash')
    try:
        response_properties = await db.create_container_if_not_exists(id=id, partition_key=partition_key, return_properties=True)
        print('Container with id \'{0}\' created or retrieved'.format(id))
        print(response_properties[1].get_response_headers())
    except exceptions.CosmosHttpResponseError as e:
        print('Error creating container: {0}'.format(e.message))


async def create_container_if_exists(db, id):
    print("\n2.3 Get Container if exists")

    try:
        response_properties = await db.create_container_if_not_exists(id=id, return_properties=True)
        print('Container with id \'{0}\' found'.format(id))
        print(response_properties[1].get_response_headers())
    except exceptions.CosmosResourceNotFoundError:
        print('Container with id \'{0}\' does not exist'.format(id))


async def read_container(db, id):
    print("\n2.4 Get a Container by id")

    try:
        container = db.get_container_client(id)
        properties = await container.read()
        print('Container with id \'{0}\' was found, it\'s link is {1}'.format(container.id, container.container_link))
    except exceptions.CosmosResourceNotFoundError:
        print('A container with id \'{0}\' does not exist'.format(id))


async def replace_throughput_container(db, id):
    print("\n2.5 Replace Container Throughput")

    try:
        container = await db.create_container(id=id,  partition_key=PartitionKey(path="/company"), offer_throughput=400)
        replace_throughput_value = 500
        properties = await container.replace_throughput(replace_throughput_value)
        # Set new throughput to 600 RU/s
        new_throughput = ThroughputProperties(offer_throughput=600)
        print('Container throughput changed to 600 RU/s')
        print(properties.get_response_headers())
    except exceptions.CosmosResourceNotFoundError:
        print('Container with id \'{0}\' does not exist'.format(id))
    except exceptions.CosmosHttpResponseError as e:
        print('Error changing container throughput: {0}'.format(e.message))


async def run_sample():
    async with CosmosClient(HOST, {'masterKey': MASTER_KEY}) as client:
        try:
            # setup database for this sample
            try:
                throughput_properties = ThroughputProperties(500)
                db = await client.create_database(id=DATABASE_ID, offer_throughput=throughput_properties)

            except exceptions.CosmosResourceExistsError:
                db = client.get_database_client(DATABASE_ID)

            # demonstrate database operations
            print("\n=== Database Operations ===")

            # create a database
            await create_db(client, DATABASE_ID + "_demo1")

            # create a database if it doesn't exist already
            await create_db_if_not_exists(client, DATABASE_ID + "_demo2")

            # read database information if the database already exists
            await create_db_if_exists(client, DATABASE_ID)

            # read from database
            await read_db(client, DATABASE_ID)

            # replace throughput for database
            await replace_throughput_db(client, DATABASE_ID)

            # demonstrate container operations
            print("\n=== Container Operations ===")

            # create a container
            await create_container(db, CONTAINER_ID)

            # create container if not exists
            await create_container_if_not_exists(db, CONTAINER_ID + "_demo")

            # get a container using its id
            await read_container(db, CONTAINER_ID)

            # replace container throughput
            await replace_throughput_container(db, CONTAINER_ID + "demo")

            # cleanup database after sample
            try:
                await client.delete_database(db)

            except exceptions.CosmosResourceNotFoundError:
                pass

        except exceptions.CosmosHttpResponseError as e:
            print('\nrun_sample has caught an error. {0}'.format(e.message))

        finally:
            print("\nrun_sample done")


if __name__ == '__main__':
    asyncio.run(run_sample())